import pandas as pd
from mindsdb_sql.parser import ast
from typing import Text, List, Dict, Tuple

from mindsdb.integrations.handlers.utilities.query_utilities.base_query_utilities import BaseQueryParser
from mindsdb.integrations.handlers.utilities.query_utilities.base_query_utilities import BaseQueryExecutor


class SELECTQueryParser(BaseQueryParser):
    """
    Parses a SELECT query into its component parts.

    Parameters
    ----------
    query : ast.Select
        Given SQL SELECT query.
    table : Text
        Name of the table to query.
    columns : List[Text]
        List of columns in the table.
    """
    def __init__(self, query: ast.Select, table: Text, columns: List[Text]):
        super().__init__(query)
        self.table = table
        self.columns = columns

    def parse_query(self) -> Tuple[List[Text], List[List[Text]], Dict[Text, List[Text]], int]:
        """
        Parses a SQL SELECT statement into its components: SELECT, WHERE, ORDER BY, LIMIT.
        """
        selected_columns = self.parse_select_clause()
        where_conditions = self.parse_where_clause()
        order_by_conditions = self.parse_order_by_clause()
        result_limit = self.parse_limit_clause()

        return selected_columns, where_conditions, order_by_conditions, result_limit

    def parse_select_clause(self) -> List[Text]:
        """
        Parses the SELECT (column selection) clause of the query.
        """
        selected_columns = []
        for target in self.query.targets:
            if isinstance(target, ast.Star):
                selected_columns = self.columns
                break
            elif isinstance(target, ast.Identifier):
                selected_columns.append(target.parts[-1])
            else:
                raise ValueError(f"Unknown query target {type(target)}")

        return selected_columns

    def parse_order_by_clause(self) -> Dict[Text, List[Text]]:
        """
        Parses the ORDER BY clause of the query.
        """
        order_by_conditions = {}
        if self.query.order_by and len(self.query.order_by) > 0:
            order_by_conditions["columns"] = []
            order_by_conditions["ascending"] = []

            for an_order in self.query.order_by:
                if an_order.field.parts[0] == self.table:
                    if an_order.field.parts[1] in self.columns:
                        order_by_conditions["columns"].append(an_order.field.parts[1])

                        if an_order.direction == "ASC":
                            order_by_conditions["ascending"].append(True)
                        else:
                            order_by_conditions["ascending"].append(False)
                    else:
                        raise ValueError(
                            f"Order by unknown column {an_order.field.parts[1]}"
                        )

        return order_by_conditions

    def parse_limit_clause(self) -> int:
        """
        Parses the LIMIT clause of the query.
        """
        if self.query.limit:
            result_limit = self.query.limit.value
        else:
            result_limit = 20

        return result_limit


class SELECTQueryExecutor(BaseQueryExecutor):
    """
    Executes a SELECT query.

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe to query.
    selected_columns : List[Text]
        List of columns to select.
    where_conditions : List[List[Text]]
        List of where conditions.
    order_by_conditions : Dict[Text, List[Text]]
        Dictionary of order by conditions.
    result_limit : int
        Number of results to return.
    """
    def __init__(self, df: pd.DataFrame, selected_columns: List[Text], where_conditions: List[List[Text]], order_by_conditions: Dict[Text, List[Text]], result_limit: int = None):
        super().__init__(df, where_conditions)
        self.selected_columns = selected_columns
        self.order_by_conditions = order_by_conditions
        self.result_limit = result_limit

    def execute_query(self):
        """
        Execute the query.
        """
        self.execute_limit_clause()

        self.execute_where_clause()

        self.execute_select_clause()

        self.execute_order_by_clause()

        return self.df

    def execute_select_clause(self):
        """
        Execute the select clause of the query.
        """
        if len(self.df) == 0:
            self.df = pd.DataFrame([], columns=self.selected_columns)
        else:
            self.df = self.df[self.selected_columns]

    def execute_order_by_clause(self):
        """
        Execute the order by clause of the query.
        """
        if len(self.order_by_conditions.get("columns", [])) > 0:
            self.df = self.df.sort_values(
                by=self.order_by_conditions["columns"],
                ascending=self.order_by_conditions["ascending"],
            )

    def execute_limit_clause(self):
        """
        Execute the limit clause of the query.
        """
        if self.result_limit:
            self.df = self.df.head(self.result_limit)


from datetime import timedelta
from collections import OrderedDict
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb_sql.parser.ast.base import ASTNode
import pandas as pd

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE,
)
from couchbase.auth import PasswordAuthenticator
from couchbase.cluster import Cluster
from couchbase.exceptions import UnAmbiguousTimeoutException
from couchbase.options import ClusterOptions
from couchbase.exceptions import KeyspaceNotFoundException, CouchbaseException


class CouchbaseHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Couchbase statements.
    """

    name = "couchbase"
    # TODO: Check the timeout value with the sdk default time
    DEFAULT_TIMEOUT_SECONDS = 60

    def __init__(self, name, **kwargs):
        super().__init__(name)
        self.connection_data = kwargs.get("connection_data")

        self.scope = self.connection_data.get("scope") or "_default"

        self.bucket_name = self.connection_data.get("bucket")
        self.cluster = None

        self.is_connected = False

    def connect(self):
        """
        Set up connections required by the handler.

        Returns:
            The connected cluster.
        """
        if self.is_connected:
            return self.cluster

        auth = PasswordAuthenticator(
            self.connection_data.get("user"),
            self.connection_data.get("password"),
            # NOTE: If using SSL/TLS, add the certificate path.
            # We strongly reccomend this for production use.
            # cert_path=cert_path
        )

        options = ClusterOptions(auth)

        if 'cloud.couchbase.com' in self.connection_data.get("host"):
            options.apply_profile('wan_development')

            endpoint = f"couchbases://{self.connection_data.get('host')}"
        else:
            endpoint = f"couchbase://{self.connection_data.get('host')}"

        cluster = Cluster(
            endpoint,
            options,
        )

        try:
            cluster.wait_until_ready(timedelta(seconds=self.DEFAULT_TIMEOUT_SECONDS))
            self.is_connected = cluster.connected
            self.cluster = cluster
        except UnAmbiguousTimeoutException:
            self.is_connected = False
            raise

        return self.cluster

    def disconnect(self):
        """Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is False:
            return
        self.is_connected = self.cluster.connected
        return

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Couchbase bucket
        :return: success status and error message if error occurs
        """
        result = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            cluster = self.connect()
            result.success = cluster.connected
        except UnAmbiguousTimeoutException as e:
            log.logger.error(
                f'Error connecting to Couchbase {self.connection_data["bucket"]}, {e}!'
            )
            result.error_message = str(e)

        if result.success is True and need_to_close:
            self.disconnect()
        if result.success is False and self.is_connected is True:
            self.is_connected = False
        return result

    def native_query(self, query: str) -> Response:
        """Execute a raw query against Couchbase.

        Args:
            query (str): Raw Couchbase query.

        Returns:
            HandlerResponse containing query results.
        """
        self.connect()
        bucket = self.cluster.bucket(self.bucket_name)
        cb = bucket.scope(self.scope)

        data = {}
        try:
            for collection in cb.query(query):
                for collection_name, row in collection.items():
                    if isinstance(row, dict):
                        for k, v in row.items():
                            data.setdefault(k, []).append(v)
                    else:
                        for k, v in collection.items():
                            data.setdefault(k, []).append(v)

            response = Response(
                RESPONSE_TYPE.TABLE, pd.DataFrame(data) if data else RESPONSE_TYPE.OK
            )
        except CouchbaseException as e:
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e.error_context.first_error_message),
            )

        return response

    def query(self, query: ASTNode) -> Response:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        return self.native_query(query.to_string())

    def get_tables(self) -> Response:
        """
        Get a list with of collection in database
        """
        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        unique_collections = set()
        for scope in bucket.collections().get_all_scopes():
            for collection in scope.collections:
                unique_collections.add(collection.name)
        collections = list(unique_collections)
        df = pd.DataFrame(collections, columns=["TABLE_NAME"])
        response = Response(RESPONSE_TYPE.TABLE, df)

        return response

    def get_columns(self, table_name) -> Response:
        """Returns a list of entity columns
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        """

        response = Response(False)

        cluster = self.connect()
        bucket = cluster.bucket(self.bucket_name)
        cb = bucket.scope(self.scope)

        try:
            q = f"SELECT * FROM `{table_name}` limit 1"
            row_iter = cb.query(q)
            # print(row_iter.execute())
            data = []
            for row in row_iter:
                for k, v in row[table_name].items():
                    data.append([k, type(v).__name__])
            df = pd.DataFrame(data, columns=["Field", "Type"])
            response = Response(RESPONSE_TYPE.TABLE, df)
        except KeyspaceNotFoundException as e:
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=f"Error: {e.error_context.first_error_message}",
            )

        return response


connection_args = OrderedDict(
    user={
        "type": ARG_TYPE.STR,
        "description": "The user name used to authenticate with the Couchbase server.",
    },
    password={
        "type": ARG_TYPE.STR,
        "description": "The password to authenticate the user with the Couchbase server.",
    },
    bucket={
        "type": ARG_TYPE.STR,
        "description": "The database/bucket name to use when connecting with the Couchbase server.",
    },
    host={
        "type": ARG_TYPE.STR,
        "description": "--your-instance--.dp.cloud.couchbase.com or IP address of the Couchbase server.",
    },
    scope={
        "type": ARG_TYPE.STR,
        "description": 'The scope use in the query context in Couchbase server. If blank, scope will be "_default".',
    },
)
connection_args_example = OrderedDict(
    host="127.0.0.1", user="root", password="password", bucket="bucket"
)



from mindsdb.utilities import log
from mindsdb.utilities.config import Config
from mindsdb.interfaces.database.integrations import integration_controller
from mindsdb.utilities.context import context as ctx


class StreamController():
    known_dbs = {}

    def __init__(self):
        self.config = Config()
        self.company_id = ctx.company_id
        self.integration_controller = integration_controller

    def setup(self, db_alias):
        try:
            integration_meta = self.integration_controller.get(db_alias)
            if integration_meta is None:
                raise Exception(f'Unkonw database integration: {db_alias}')
            if integration_meta.get('engine') not in self.known_dbs:
                raise Exception(f'Unkonw database integration type for: {db_alias}')
            self.known_dbs[integration_meta['engine']](self.config, db_alias, integration_meta).setup()
        except Exception as e:
            log.logger.warning('Failed to setup stream for ' + db_alias + f', error: {e}')



from mindsdb_sql.parser.ast import Describe, Identifier

from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers
from mindsdb.api.mongo.classes.query_sql import run_sql_command


class Responce(Responder):
    when = {'collStats': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        db = query['$db']
        collection = query['collStats']

        scale = query.get('scale')

        if db != 'mindsdb' or collection == 'predictors' or scale is None:
            # old behavior
            # NOTE real answer is huge, i removed most data from it.
            res = {
                'ns': "db.collection",
                'size': 1,
                'count': 0,
                'avgObjSize': 1,
                'storageSize': 16384,
                'capped': False,
                'wiredTiger': {
                },
                'nindexes': 1,
                'indexDetails': {
                },
                'totalIndexSize': 16384,
                'indexSizes': {
                    '_id_': 16384
                },
                'ok': 1
            }

            res['ns'] = f"{db}.{collection}"
            if db == 'mindsdb' and collection == 'predictors':
                res['count'] = len(mindsdb_env['model_controller'].get_models())
        else:

            ident_parts = [collection]
            if scale is not None:
                ident_parts.append(scale)

            ast_query = Describe(Identifier(
                parts=ident_parts
            ))

            data = run_sql_command(request_env, ast_query)
            res = {
                'data': data
            }

        res['ns'] = f"{db}.{collection}"

        return res


responder = Responce()



from typing import Optional
from collections import OrderedDict

import pandas as pd
import boto3
import io
import ast

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class S3Handler(DatabaseHandler):
    """
    This handler handles connection and execution of the S3 statements.
    """

    name = 's3'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 's3'
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = boto3.client(
            's3',
            aws_access_key_id=self.connection_data['aws_access_key_id'],
            aws_secret_access_key=self.connection_data['aws_secret_access_key'],
            region_name=self.connection_data['region_name']
        )
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """ Close any existing connections
        Should switch self.is_connected.
        """
        self.is_connected = False
        return

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            connection.head_object(Bucket=self.connection_data['bucket'], Key=self.connection_data['key'])
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to AWS with the given credentials, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()

        try:
            result = connection.select_object_content(
                Bucket=self.connection_data['bucket'],
                Key=self.connection_data['key'],
                ExpressionType='SQL',
                Expression=query,
                InputSerialization=ast.literal_eval(self.connection_data['input_serialization']),
                OutputSerialization={"CSV": {}}
            )

            records = []
            for event in result['Payload']:
                if 'Records' in event:
                    records.append(event['Records']['Payload'])
                elif 'Stats' in event:
                    stats = event['Stats']['Details']

            file_str = ''.join(r.decode('utf-8') for r in records)

            df = pd.read_csv(io.StringIO(file_str))

            response = Response(
                RESPONSE_TYPE.TABLE,
                data_frame=df
            )
        except Exception as e:
            log.logger.error(f'Error running query: {query} on {self.connection_data["key"]} in {self.connection_data["bucket"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """

        return self.native_query(query.to_string())

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        connection = self.connect()
        objects = [obj['Key'] for obj in connection.list_objects(Bucket=self.connection_data["bucket"])['Contents']]

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                objects,
                columns=['table_name']
            )
        )

        return response

    def get_columns(self) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = "SELECT * FROM S3Object LIMIT 5"
        result = self.native_query(query)

        response = Response(
            RESPONSE_TYPE.TABLE,
            data_frame=pd.DataFrame(
                {
                    'column_name': result.data_frame.columns,
                    'data_type': result.data_frame.dtypes
                }
            )
        )

        return response


connection_args = OrderedDict(
    aws_access_key_id={
        'type': ARG_TYPE.STR,
        'description': 'The access key for the AWS account.'
    },
    aws_secret_access_key={
        'type': ARG_TYPE.STR,
        'description': 'The secret key for the AWS account.'
    },
    region_name={
        'type': ARG_TYPE.STR,
        'description': 'The AWS region where the S3 bucket is located.'
    },
    bucket={
        'type': ARG_TYPE.STR,
        'description': 'The name of the S3 bucket.'
    },
    key={
        'type': ARG_TYPE.STR,
        'description': 'The key of the object to be queried.'
    },
    input_serialization={
        'type': ARG_TYPE.STR,
        'description': 'The format of the data in the object that is to be queried.'
    }
)

connection_args_example = OrderedDict(
    aws_access_key_id='PCAQ2LJDOSWLNSQKOCPW',
    aws_secret_access_key='U/VjewPlNopsDmmwItl34r2neyC6WhZpUiip57i',
    region_name='us-east-1',
    bucket='mindsdb-bucket',
    key='iris.csv',
    input_serialization="{'CSV': {'FileHeaderInfo': 'NONE'}}",
)





import json
import os

from flask import Response


def http_error(status_code, title, detail=''):
    ''' Wrapper for error responce acoording with RFC 7807 (https://tools.ietf.org/html/rfc7807)

        :param status_code: int - http status code for response
        :param title: str
        :param detail: str

        :return: flask Response object
    '''
    return Response(
        response=json.dumps({
            'title': title,
            'detail': detail
        }),
        status=status_code,
        headers={
            'Content-Type': 'application/problem+json'
        }
    )


def __is_within_directory(directory, target):
    abs_directory = os.path.abspath(directory)
    abs_target = os.path.abspath(target)
    prefix = os.path.commonprefix([abs_directory, abs_target])
    return prefix == abs_directory


def safe_extract(tar, path=".", members=None, *, numeric_owner=False):
    for member in tar.getmembers():
        member_path = os.path.join(path, member.name)
        if not __is_within_directory(path, member_path):
            raise Exception("Attempted Path Traversal in Tar File")
    tar.extractall(path, members, numeric_owner)



OPENAI_API_BASE = 'https://api.openai.com/v1'

CHAT_MODELS = ('gpt-3.5-turbo', 'gpt-3.5-turbo-16k', 'gpt-3.5-turbo-instruct', 'gpt-4', 'gpt-4-32k')
COMPLETION_MODELS = ('babbage-002', 'davinci-002')
FINETUNING_MODELS = ('gpt-3.5-turbo', 'babbage-002', 'davinci-002')
COMPLETION_LEGACY_BASE_MODELS = ('davinci', 'curie', 'babbage', 'ada')

FINETUNING_LEGACY_MODELS = COMPLETION_LEGACY_BASE_MODELS
COMPLETION_LEGACY_MODELS = COMPLETION_LEGACY_BASE_MODELS + \
                        tuple(f'text-{model}-001' for model in COMPLETION_LEGACY_BASE_MODELS) + \
                        ('text-davinci-002', 'text-davinci-003')

EMBEDDING_MODELS = ('text-embedding-ada-002',) + \
                tuple(f'text-similarity-{model}-001' for model in COMPLETION_LEGACY_BASE_MODELS) + \
                tuple(f'text-search-{model}-query-001' for model in COMPLETION_LEGACY_BASE_MODELS) + \
                tuple(f'text-search-{model}-doc-001' for model in COMPLETION_LEGACY_BASE_MODELS) + \
                tuple(f'code-search-{model}-text-001' for model in COMPLETION_LEGACY_BASE_MODELS) + \
                tuple(f'code-search-{model}-code-001' for model in COMPLETION_LEGACY_BASE_MODELS)

ALL_MODELS = list(set(CHAT_MODELS + COMPLETION_MODELS + COMPLETION_LEGACY_MODELS + EMBEDDING_MODELS))  # noqa


import ast
import inspect
import logging
import textwrap
from _ast import AnnAssign, AugAssign
from typing import Any, Dict, List, Optional

import pandas as pd
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.integrations.libs.response import HandlerResponse, HandlerStatusResponse

LOG = logging.getLogger(__name__)


class BaseHandler:
    """ Base class for database handlers

    Base class for handlers that associate a source of information with the
    broader MindsDB ecosystem via SQL commands.
    """

    def __init__(self, name: str):
        """ constructor
        Args:
            name (str): the handler name
        """
        self.is_connected: bool = False
        self.name = name

    def connect(self):
        """ Set up any connections required by the handler

        Should return connection

        """
        raise NotImplementedError()

    def disconnect(self):
        """ Close any existing connections

        Should switch self.is_connected.
        """
        self.is_connected = False
        return

    def check_connection(self) -> HandlerStatusResponse:
        """ Check connection to the handler

        Returns:
            HandlerStatusResponse
        """
        raise NotImplementedError()

    def native_query(self, query: Any) -> HandlerResponse:
        """Receive raw query and act upon it somehow.

        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def query(self, query: ASTNode) -> HandlerResponse:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.

        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INSERT, DELETE, etc

        Returns:
            HandlerResponse
        """
        raise NotImplementedError()

    def get_tables(self) -> HandlerResponse:
        """ Return list of entities

        Return list of entities that will be accesible as tables.

        Returns:
            HandlerResponse: shoud have same columns as information_schema.tables
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-tables-table.html)
                Column 'TABLE_NAME' is mandatory, other is optional.
        """
        raise NotImplementedError()

    def get_columns(self, table_name: str) -> HandlerResponse:
        """ Returns a list of entity columns

        Args:
            table_name (str): name of one of tables returned by self.get_tables()

        Returns:
            HandlerResponse: shoud have same columns as information_schema.columns
                (https://dev.mysql.com/doc/refman/8.0/en/information-schema-columns-table.html)
                Column 'COLUMN_NAME' is mandatory, other is optional. Hightly
                recomended to define also 'DATA_TYPE': it should be one of
                python data types (by default it str).
        """
        raise NotImplementedError()


class DatabaseHandler(BaseHandler):
    """
    Base class for handlers associated to data storage systems (e.g. databases, data warehouses, streaming services, etc.)
    """

    def __init__(self, name: str):
        super().__init__(name)


class PredictiveHandler(BaseHandler):
    """
    DEPRECATED. Please refer to BaseMLEngine for integrations with machine learning frameworks.

    Base class for handlers associated to predictive systems.
    """

    def __init__(self, name: str):
        super().__init__(name)


class ArgProbeMixin:
    """
    A mixin class that provides probing of arguments that
    are needed by a handler during creation and prediction time
    by running the static analysis on the source code of the handler.
    """

    class ArgProbeVisitor(ast.NodeVisitor):
        def __init__(self):
            self.arg_keys = []
            self.var_names_to_track = {"args"}

        def visit_Assign(self, node):
            # track if args['using'] get assigned to any variable
            # if so, we should track the variable by adding it to
            # self.var_names_to_track
            # E.g., using_args = args['using']
            # we should track using_args as well
            if (
                isinstance(node.value, ast.Subscript)
                and isinstance(node.value.value, ast.Name)
                and node.value.value.id == "args"
            ):
                if (
                    isinstance(node.value.slice, ast.Index)
                    and isinstance(node.value.slice.value, ast.Str)
                    and node.value.slice.value.s == "using"
                ):
                    self.var_names_to_track.add(node.targets[0].id)

            # for an assignment like `self.args['name'] = 'value'`, we should ignore
            # the left side of the assignment
            self.visit(node.value)

        def visit_AnnAssign(self, node: AnnAssign) -> Any:
            self.visit(node.value)

        def visit_AugAssign(self, node: AugAssign) -> Any:
            self.visit(node.value)

        def visit_Subscript(self, node):
            if (
                isinstance(node.value, ast.Name)
                and node.value.id in self.var_names_to_track
            ):
                if isinstance(node.slice, ast.Index) and isinstance(
                    node.slice.value, ast.Str
                ):
                    self.arg_keys.append({"name": node.slice.value.s, "required": True})
            self.generic_visit(node)

        def visit_Call(self, node):
            if isinstance(node.func, ast.Attribute) and node.func.attr == "get":
                if (
                    isinstance(node.func.value, ast.Name)
                    and node.func.value.id in self.var_names_to_track
                ):
                    if isinstance(node.args[0], ast.Str):
                        self.arg_keys.append(
                            {"name": node.args[0].s, "required": False}
                        )
            self.generic_visit(node)

    @classmethod
    def probe_function(self, method_name: str) -> List[Dict]:
        """
        Probe the source code of the method with name method_name.
        Specifically, trace how the argument `args`, which is a dict is used in the method.

        Find all places where a key of the dict is used, and return a list of all keys that are used.
        E.g.,
        args["key1"] -> "key1" is accessed, and it is required
        args.get("key2", "default_value") -> "key2" is accessed, and it is optional (default value is provided)

        Return a list of dict
        where each dict looks like
        {
            "name": "key1",
            "required": True
        }
        """
        try:
            source_code = self.get_source_code(method_name)
        except Exception as e:
            LOG.error(
                f"Failed to get source code of method {method_name} in {self.__class__.__name__}. Reason: {e}"
            )
            return []

        # parse the source code
        # fix the indentation
        source_code = textwrap.dedent(source_code)
        # parse the source code
        tree = ast.parse(source_code)

        # find all places where a key in args is accessed
        # and if it is accessed using args["key"] or args.get("key", "default_value")

        visitor = self.ArgProbeVisitor()
        visitor.visit(tree)

        # deduplicate the keys
        # if there two records with the same name but different required status
        # we should keep the one with required == True
        unique_arg_keys = {}
        for r in visitor.arg_keys:
            if r["name"] in unique_arg_keys:
                if r["required"]:
                    unique_arg_keys[r["name"]] = r["required"]
            else:
                unique_arg_keys[r["name"]] = r["required"]

        # convert back to list
        visitor.arg_keys = [{"name": k, "required": v} for k, v in unique_arg_keys.items()]

        # filter out record where name == "using"
        return [r for r in visitor.arg_keys if r["name"] != "using"]

    @classmethod
    def get_source_code(self, method_name: str):
        """
        Get the source code of the method specified by method_name
        """
        method = getattr(self, method_name)
        if method is None:
            raise Exception(
                f"Method {method_name} does not exist in {self.__class__.__name__}"
            )
        source_code = inspect.getsource(method)
        return source_code

    @classmethod
    def prediction_args(self):
        """
        Get the arguments that are needed by the prediction method
        """
        return self.probe_function("predict")

    @classmethod
    def creation_args(self):
        """
        Get the arguments that are needed by the creation method
        """
        return self.probe_function("create")


class BaseMLEngine(ArgProbeMixin):
    """
    Base class for integration engines to connect with other machine learning libraries/frameworks.

    This class will be instanced when interacting with the underlying framework. For compliance with the interface
    that MindsDB core expects, instances of this class will be wrapped with the `BaseMLEngineExec` class defined
    in `libs/ml_exec_base`.

    Broadly speaking, the flow is as follows:
      - A SQL statement is sent to the MindsDB executor
      - The statement is parsed, and a sequential plan is generated by `mindsdb_sql`
      - If any step in the plan involves an ML framework, a wrapped engine that inherits from this class will be called for the respective action
          - For example, creating a new model would call `create()`
      - Any output produced by the ML engine is then formatted by the wrapper and passed back into the MindsDB executor, which can then morph the data to comply with the original SQL query
    """  # noqa

    def __init__(self, model_storage, engine_storage, **kwargs) -> None:
        """
        Warning: This method should not be overridden.

        Initialize storage objects required by the ML engine.

        - engine_storage: persists global engine-related internals or artifacts that may be used by all models from the engine.
        - model_storage: stores artifacts for any single given model.
        """
        self.model_storage = model_storage
        self.engine_storage = engine_storage
        self.generative = False  # if True, the target column name does not have to be specified at creation time

        if kwargs.get('base_model_storage'):
            self.base_model_storage = kwargs['base_model_storage']  # available when updating a model
        else:
            self.base_model_storage = None

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """
        Saves a model inside the engine registry for later usage.

        Normally, an input dataframe is required to train the model.
        However, some integrations may merely require registering the model instead of training, in which case `df` can be omitted.

        Any other arguments required to register the model can be passed in an `args` dictionary.
        """
        raise NotImplementedError

    def predict(self, df: pd.DataFrame, args: Optional[Dict] = None) -> pd.DataFrame:
        """
        Calls a model with some input dataframe `df`, and optionally some arguments `args` that may modify the model behavior.

        The expected output is a dataframe with the predicted values in the target-named column.
        Additional columns can be present, and will be considered row-wise explanations if their names finish with `_explain`.
        """
        raise NotImplementedError

    def finetune(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        """
        Optional.

        Used to fine-tune a pre-existing model without resetting its internal state (e.g. weights).

        Availability will depend on underlying integration support, as not all ML models can be partially updated.
        """
        raise NotImplementedError

    def describe(self, attribute: Optional[str] = None) -> pd.DataFrame:
        """
        Optional.

        When called, this method provides global model insights, e.g. framework-level parameters used in training.
        """
        raise NotImplementedError

    def create_engine(self, connection_args: dict):
        """
        Optional.

        Used to connect with external sources (e.g. a REST API) that the engine will require to use any other methods.
        """
        raise NotImplementedError

    def close(self):
        pass



"""training_data_rows_columns_count

Revision ID: b5b53e0ea7f8
Revises: 999bceb904df
Create Date: 2022-07-15 12:21:40.523039

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db    # noqa


# revision identifiers, used by Alembic.
revision = 'b5b53e0ea7f8'
down_revision = '999bceb904df'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('training_data_columns_count', sa.Integer(), nullable=True))
        batch_op.add_column(sa.Column('training_data_rows_count', sa.Integer(), nullable=True))


def downgrade():
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('training_data_rows_count')
        batch_op.drop_column('training_data_columns_count')



from collections import OrderedDict

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from ..mysql_handler import Handler as MySQLHandler



class EdgelessDBHandler(MySQLHandler):
    """
    This handler handles connection and execution of the EdgelessDB statements.
    """
    name = 'edgelessdb'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the EdgelessDB server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the EdgelessDB server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the EdgelessDB server.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the EdgelessDB server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the EdgelessDB server. Must be an integer.'
    },
    ssl={
        'type': ARG_TYPE.BOOL,
        'description': 'Set it to False to disable ssl.'
    },
    ssl_ca={
        'type': ARG_TYPE.PATH,
        'description': 'Path or URL of the Certificate Authority (CA) certificate file'
    },
    ssl_cert={
        'type': ARG_TYPE.PATH,
        'description': 'Path name or URL of the server public key certificate file'
    },
    ssl_key={
        'type': ARG_TYPE.PATH,
        'description': 'The path name or URL of the server private key file'
    }
)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=3306,
    user='root',
    password='password',
    database='database'
)



import time
import importlib
import traceback
import datetime as dt
from collections import UserDict

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import Identifier, Select, Star, NativeQuery

import mindsdb.interfaces.storage.db as db
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import SQLQuery
from mindsdb.integrations.utilities.sql_utils import make_sql_session
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage
from mindsdb.interfaces.model.functions import get_model_records
from mindsdb.integrations.utilities.utils import format_exception_error
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.functions import mark_process
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.config import Config


class HandlersCache(UserDict):
    def __init__(self, max_size: int = 5) -> None:
        self._max_size = max_size
        super().__init__()

    def __setitem__(self, key, value) -> None:
        if len(self.data) > self._max_size:
            sorted_elements = sorted(
                self.data.items(),
                key=lambda x: x[1]['last_usage_at']
            )
            del self.data[sorted_elements[0][0]]
        self.data[key] = {
            'last_usage_at': time.time(),
            'handler': value
        }

    def __getitem__(self, key: int) -> object:
        el = super().__getitem__(key)
        el['last_usage_at'] = time.time()
        return el['handler']


handlers_cacher = HandlersCache()


@mark_process(name='learn')
def predict_process(payload, dataframe):
    db.init()
    integration_id = payload['handler_meta']['integration_id']
    predictor_record = payload['predictor_record']
    args = payload['args']
    module_path = payload['handler_meta']['module_path']
    class_name = payload['handler_meta']['class_name']
    ml_engine_name = payload['handler_meta']['engine']

    module = importlib.import_module(module_path)
    HandlerClass = getattr(module, class_name)
    handler_class = HandlerClass

    if predictor_record.id not in handlers_cacher:
        handlerStorage = HandlerStorage(integration_id)
        modelStorage = ModelStorage(predictor_record.id)
        ml_handler = handler_class(
            engine_storage=handlerStorage,
            model_storage=modelStorage,
        )
        handlers_cacher[predictor_record.id] = ml_handler
    else:
        ml_handler = handlers_cacher[predictor_record.id]

    if ml_engine_name == 'lightwood':
        args['code'] = predictor_record.code
        args['target'] = predictor_record.to_predict[0]
        args['dtype_dict'] = predictor_record.dtype_dict
        args['learn_args'] = predictor_record.learn_args

    if ml_engine_name == 'langchain':
        from mindsdb.api.mysql.mysql_proxy.controllers import SessionController
        from mindsdb.api.mysql.mysql_proxy.executor.executor_commands import ExecuteCommands

        sql_session = SessionController()
        sql_session.database = 'mindsdb'

        command_executor = ExecuteCommands(sql_session, executor=None)

        args['executor'] = command_executor

    predictions = ml_handler.predict(dataframe, args)
    ml_handler.close()
    return predictions


@mark_process(name='learn')
def learn_process(payload, dataframe):
    ctx.profiling = {
        'level': 0,
        'enabled': True,
        'pointer': None,
        'tree': None
    }
    profiler.set_meta(query='learn_process', api='http', environment=Config().get('environment'))
    with profiler.Context('learn_process'):
        from mindsdb.interfaces.database.database import DatabaseController
        db.init()

        data_integration_ref = payload['data_integration_ref']
        problem_definition = payload['problem_definition']
        fetch_data_query = payload['fetch_data_query']
        project_name = payload['project_name']
        predictor_id = payload['model_id']
        # engine = payload['handler_meta']['engine']
        integration_id = payload['handler_meta']['integration_id']
        base_predictor_id = payload.get('base_model_id')
        set_active = payload['set_active']
        class_path = (payload['handler_meta']['module_path'], payload['handler_meta']['class_name'])

        try:
            target = problem_definition.get('target', None)
            training_data_df = None
            if data_integration_ref is not None:
                database_controller = DatabaseController()
                sql_session = make_sql_session()
                if data_integration_ref['type'] == 'integration':
                    integration_name = database_controller.get_integration(data_integration_ref['id'])['name']
                    query = Select(
                        targets=[Star()],
                        from_table=NativeQuery(
                            integration=Identifier(integration_name),
                            query=fetch_data_query
                        )
                    )
                    sqlquery = SQLQuery(query, session=sql_session)
                elif data_integration_ref['type'] == 'view':
                    project = database_controller.get_project(project_name)
                    query_ast = parse_sql(fetch_data_query, dialect='mindsdb')
                    view_meta = project.query_view(query_ast)
                    sqlquery = SQLQuery(view_meta['query_ast'], session=sql_session)

                result = sqlquery.fetch(view='dataframe')
                training_data_df = result['result']

            training_data_columns_count, training_data_rows_count = 0, 0
            if training_data_df is not None:
                training_data_columns_count = len(training_data_df.columns)
                training_data_rows_count = len(training_data_df)

            predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
            predictor_record.training_data_columns_count = training_data_columns_count
            predictor_record.training_data_rows_count = training_data_rows_count
            db.session.commit()

            module_name, class_name = class_path
            module = importlib.import_module(module_name)
            HandlerClass = getattr(module, class_name)

            handlerStorage = HandlerStorage(integration_id)
            modelStorage = ModelStorage(predictor_id)
            modelStorage.fileStorage.push()     # FIXME

            kwargs = {}
            if base_predictor_id is not None:
                kwargs['base_model_storage'] = ModelStorage(base_predictor_id)
                kwargs['base_model_storage'].fileStorage.pull()

            ml_handler = HandlerClass(
                engine_storage=handlerStorage,
                model_storage=modelStorage,
                **kwargs
            )
            handlers_cacher[predictor_record.id] = ml_handler

            if not ml_handler.generative:
                if training_data_df is not None and target not in training_data_df.columns:
                    raise Exception(
                        f'Prediction target "{target}" not found in training dataframe: {list(training_data_df.columns)}')

            # create new model
            if base_predictor_id is None:
                with profiler.Context('create'):
                    ml_handler.create(target, df=training_data_df, args=problem_definition)

            # fine-tune (partially train) existing model
            else:
                # load model from previous version, use it as starting point
                with profiler.Context('finetune'):
                    problem_definition['base_model_id'] = base_predictor_id
                    ml_handler.finetune(df=training_data_df, args=problem_definition)

            predictor_record.status = PREDICTOR_STATUS.COMPLETE
            predictor_record.active = set_active
            db.session.commit()
            # if retrain and set_active after success creation
            if set_active is True:
                models = get_model_records(
                    name=predictor_record.name,
                    project_id=predictor_record.project_id,
                    active=None
                )
                for model in models:
                    model.active = False
                models = [x for x in models if x.status == PREDICTOR_STATUS.COMPLETE]
                models.sort(key=lambda x: x.created_at)
                models[-1].active = True
        except Exception as e:
            print(traceback.format_exc())
            error_message = format_exception_error(e)

            predictor_record = db.Predictor.query.with_for_update().get(predictor_id)
            predictor_record.data = {"error": error_message}
            predictor_record.status = PREDICTOR_STATUS.ERROR
            db.session.commit()

        predictor_record.training_stop_at = dt.datetime.now()
        db.session.commit()



from mindsdb.integrations.libs.const import HANDLER_TYPE

from .byom_handler import BYOMHandler as Handler, connection_args
from .__about__ import __version__ as version, __description__ as description

title = 'BYOM'
name = 'byom'
type = HANDLER_TYPE.ML

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'description', 'connection_args'
]




from mindsdb.integrations.libs.api_handler import APITable
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb_sql.parser import ast

from typing import Dict, List

import pandas as pd


class OpenBBtable(APITable):
    def _get_params_from_conditions(self, conditions: List) -> Dict:
        """Gets aggregate trade data API params from SQL WHERE conditions.

        Returns params to use for Binance API call to klines.

        Args:
            conditions (List): List of individual SQL WHERE conditions.
        """
        params: dict = {}
        # generic interpreter for conditions
        # since these are all equality conditions due to OpenBB Platform's API
        # then we can just use the first arg as the key and the second as the value
        for op, arg1, arg2 in conditions:
            if op != "=":
                raise NotImplementedError
            params[arg1] = arg2

        return params

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Selects data from the OpenBB Platform and returns it as a pandas DataFrame.

        Returns dataframe representing the OpenBB data.

        Args:
            query (ast.Select): Given SQL SELECT query
        """
        conditions = extract_comparison_conditions(query.where)
        params = self._get_params_from_conditions(conditions)

        openbb_data = self.handler.call_openbb_api(
            method_name="openbb_fetcher",
            params=params,
        )

        return openbb_data



import setuptools

about = {}
with open("__about__.py") as fp:
    exec(fp.read(), about)

with open("requirements.txt") as req_file:
    requirements = [req.strip() for req in req_file.read().splitlines()]

setuptools.setup(
    name=about["__title__"],
    version=about["__version__"],
    url=about["__github__"],
    download_url=about["__pypi__"],
    license=about["__license__"],
    author=about["__author__"],
    description=about["__description__"],
    packages=setuptools.find_packages(),
    package_data={"project": ["requirements.txt"]},
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



import datetime as dt
import ast
import pytz

import pandas as pd

from mindsdb_sql.parser import ast

from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.handlers.utilities.query_utilities import SELECTQueryParser, SELECTQueryExecutor
from mindsdb.integrations.handlers.utilities.query_utilities.insert_query_utilities import INSERTQueryParser


class EmailsTable(APITable):
    """The Emails Table implementation"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls email data from the connected account.

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Emails matching the query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'emails',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        search_params = {}
        for op, arg1, arg2 in where_conditions:
            if arg1 == 'created_at':
                date = self.parse_date(arg2)
                if op == '>':
                    search_params['since_date'] = date
                elif op == '<':
                    search_params['until_date'] = date
                else:
                    raise NotImplementedError("Only > and < operators are supported for created_at column.")
                continue

            elif arg1 in ['mailbox', 'subject', 'to', 'from', 'since_emailid']:
                if op != '=':
                    raise NotImplementedError("Only = operator is supported for mailbox, subject, to and from columns.")
                else:
                    if arg1 == 'from':
                        search_params['from_'] = arg2
                    else:
                        search_params[arg1] = arg2

            else:
                raise NotImplementedError(f"Unsupported column: {arg1}.")

        connection = self.handler.connect()
        emails_df = connection.search_email(**search_params)

        select_statement_executor = SELECTQueryExecutor(
            emails_df,
            selected_columns,
            [],
            order_by_conditions,
            result_limit
        )
        emails_df = select_statement_executor.execute_query()

        return emails_df

    def insert(self, query: ast.Insert) -> None:
        """Sends emails through the connected account.

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        insert_statement_parser = INSERTQueryParser(
            query,
            supported_columns=['to', 'subject', 'body'],
            mandatory_columns=['to', 'subject', 'body'],
            all_mandatory=True
        )
        email_data = insert_statement_parser.parse_query()

        for email in email_data:
            connection = self.handler.connect()
            to_addr = email['to']
            del email['to']
            connection.send_email(to_addr, **email)

    def get_columns(self):
        return ['id', 'created_at', 'to', 'from', 'subject', 'body']

    @staticmethod
    def parse_date(date_str):
        if isinstance(date_str, dt.datetime):
            return date_str
        date_formats = ['%Y-%m-%d %H:%M:%S', '%Y-%m-%d']
        date = None
        for date_format in date_formats:
            try:
                date = dt.datetime.strptime(date_str, date_format)
            except ValueError:
                pass
        if date is None:
            raise ValueError(f"Can't parse date: {date_str}")
        date = date.astimezone(pytz.utc)
        return date


# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

import mindsdb.microservices_grpc.executor.executor_pb2 as executor__pb2


class ExecutorServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.StatementPrepare = channel.unary_unary(
                '/ExecutorService/StatementPrepare',
                request_serializer=executor__pb2.ExecutionContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorResponse.FromString,
                )
        self.StatementExecute = channel.unary_unary(
                '/ExecutorService/StatementExecute',
                request_serializer=executor__pb2.StatementExecuteContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorResponse.FromString,
                )
        self.QueryExecute = channel.unary_unary(
                '/ExecutorService/QueryExecute',
                request_serializer=executor__pb2.ExecutionContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorResponse.FromString,
                )
        self.BinaryQueryExecute = channel.unary_unary(
                '/ExecutorService/BinaryQueryExecute',
                request_serializer=executor__pb2.BinaryExecutionContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorResponse.FromString,
                )
        self.ExecuteExternal = channel.unary_unary(
                '/ExecutorService/ExecuteExternal',
                request_serializer=executor__pb2.ExecutionContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorResponse.FromString,
                )
        self.Parse = channel.unary_unary(
                '/ExecutorService/Parse',
                request_serializer=executor__pb2.ExecutionContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorResponse.FromString,
                )
        self.DoExecute = channel.unary_unary(
                '/ExecutorService/DoExecute',
                request_serializer=executor__pb2.ExecutorContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorResponse.FromString,
                )
        self.ChangeDefaultDB = channel.unary_unary(
                '/ExecutorService/ChangeDefaultDB',
                request_serializer=executor__pb2.DefaultDBContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorResponse.FromString,
                )
        self.DeleteExecutor = channel.unary_unary(
                '/ExecutorService/DeleteExecutor',
                request_serializer=executor__pb2.ExecutorContext.SerializeToString,
                response_deserializer=executor__pb2.ExecutorStatusResponse.FromString,
                )


class ExecutorServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def StatementPrepare(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StatementExecute(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def QueryExecute(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def BinaryQueryExecute(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ExecuteExternal(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Parse(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DoExecute(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ChangeDefaultDB(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteExecutor(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ExecutorServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'StatementPrepare': grpc.unary_unary_rpc_method_handler(
                    servicer.StatementPrepare,
                    request_deserializer=executor__pb2.ExecutionContext.FromString,
                    response_serializer=executor__pb2.ExecutorResponse.SerializeToString,
            ),
            'StatementExecute': grpc.unary_unary_rpc_method_handler(
                    servicer.StatementExecute,
                    request_deserializer=executor__pb2.StatementExecuteContext.FromString,
                    response_serializer=executor__pb2.ExecutorResponse.SerializeToString,
            ),
            'QueryExecute': grpc.unary_unary_rpc_method_handler(
                    servicer.QueryExecute,
                    request_deserializer=executor__pb2.ExecutionContext.FromString,
                    response_serializer=executor__pb2.ExecutorResponse.SerializeToString,
            ),
            'BinaryQueryExecute': grpc.unary_unary_rpc_method_handler(
                    servicer.BinaryQueryExecute,
                    request_deserializer=executor__pb2.BinaryExecutionContext.FromString,
                    response_serializer=executor__pb2.ExecutorResponse.SerializeToString,
            ),
            'ExecuteExternal': grpc.unary_unary_rpc_method_handler(
                    servicer.ExecuteExternal,
                    request_deserializer=executor__pb2.ExecutionContext.FromString,
                    response_serializer=executor__pb2.ExecutorResponse.SerializeToString,
            ),
            'Parse': grpc.unary_unary_rpc_method_handler(
                    servicer.Parse,
                    request_deserializer=executor__pb2.ExecutionContext.FromString,
                    response_serializer=executor__pb2.ExecutorResponse.SerializeToString,
            ),
            'DoExecute': grpc.unary_unary_rpc_method_handler(
                    servicer.DoExecute,
                    request_deserializer=executor__pb2.ExecutorContext.FromString,
                    response_serializer=executor__pb2.ExecutorResponse.SerializeToString,
            ),
            'ChangeDefaultDB': grpc.unary_unary_rpc_method_handler(
                    servicer.ChangeDefaultDB,
                    request_deserializer=executor__pb2.DefaultDBContext.FromString,
                    response_serializer=executor__pb2.ExecutorResponse.SerializeToString,
            ),
            'DeleteExecutor': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteExecutor,
                    request_deserializer=executor__pb2.ExecutorContext.FromString,
                    response_serializer=executor__pb2.ExecutorStatusResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'ExecutorService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ExecutorService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def StatementPrepare(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/StatementPrepare',
            executor__pb2.ExecutionContext.SerializeToString,
            executor__pb2.ExecutorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StatementExecute(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/StatementExecute',
            executor__pb2.StatementExecuteContext.SerializeToString,
            executor__pb2.ExecutorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def QueryExecute(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/QueryExecute',
            executor__pb2.ExecutionContext.SerializeToString,
            executor__pb2.ExecutorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def BinaryQueryExecute(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/BinaryQueryExecute',
            executor__pb2.BinaryExecutionContext.SerializeToString,
            executor__pb2.ExecutorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ExecuteExternal(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/ExecuteExternal',
            executor__pb2.ExecutionContext.SerializeToString,
            executor__pb2.ExecutorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Parse(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/Parse',
            executor__pb2.ExecutionContext.SerializeToString,
            executor__pb2.ExecutorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DoExecute(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/DoExecute',
            executor__pb2.ExecutorContext.SerializeToString,
            executor__pb2.ExecutorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ChangeDefaultDB(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/ChangeDefaultDB',
            executor__pb2.DefaultDBContext.SerializeToString,
            executor__pb2.ExecutorResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeleteExecutor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/ExecutorService/DeleteExecutor',
            executor__pb2.ExecutorContext.SerializeToString,
            executor__pb2.ExecutorStatusResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from setuptools import find_packages, setup

from mindsdb.integrations.handlers.autosklearn_handler.__about__ import (
    __author__,
    __description__,
    __github__,
    __license__,
    __pypi__,
    __title__,
    __version__,
)

with open("mindsdb/lightfm_handler/requirements.txt") as req_file:
    requirements = [req.strip() for req in req_file.read().splitlines()]

setup(
    name=__title__,
    version=__version__,
    url=__github__,
    download_url=__pypi__,
    license=__license__,
    author=__author__,
    description=__description__,
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)



import traceback

from flask_restx import Resource
from flask import request

from mindsdb.api.http.namespaces.configs.sql import ns_conf
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE as SQL_RESPONSE_TYPE
from mindsdb.api.mysql.mysql_proxy.utilities import (
    SqlApiException,
    SqlApiUnknownError
)
import mindsdb.utilities.hooks as hooks
from mindsdb.utilities.context import context as ctx
import mindsdb.utilities.profiler as profiler
from mindsdb.utilities.config import Config


@ns_conf.route('/query')
@ns_conf.param('query', 'Execute query')
class Query(Resource):
    @ns_conf.doc('query')
    def post(self):
        query = request.json['query']
        context = request.json.get('context', {})

        if context.get('profiling') is True:
            profiler.enable()

        error_type = None
        error_code = None
        error_text = None
        error_traceback = None

        profiler.set_meta(query=query, api='http', environment=Config().get('environment'))
        with profiler.Context('http_query_processing'):
            mysql_proxy = FakeMysqlProxy()
            mysql_proxy.set_context(context)
            try:
                result = mysql_proxy.process_query(query)

                if result.type == SQL_RESPONSE_TYPE.OK:
                    query_response = {
                        'type': SQL_RESPONSE_TYPE.OK
                    }
                elif result.type == SQL_RESPONSE_TYPE.TABLE:
                    query_response = {
                        'type': SQL_RESPONSE_TYPE.TABLE,
                        'data': result.data,
                        'column_names': [x['alias'] or x['name'] if 'alias' in x else x['name'] for x in result.columns]
                    }
            except SqlApiException as e:
                # classified error
                error_type = 'expected'
                query_response = {
                    'type': SQL_RESPONSE_TYPE.ERROR,
                    'error_code': e.err_code,
                    'error_message': str(e)
                }

            except SqlApiUnknownError as e:
                # unclassified
                error_type = 'unexpected'
                query_response = {
                    'type': SQL_RESPONSE_TYPE.ERROR,
                    'error_code': e.err_code,
                    'error_message': str(e)
                }

            except Exception as e:
                error_type = 'unexpected'
                query_response = {
                    'type': SQL_RESPONSE_TYPE.ERROR,
                    'error_code': 0,
                    'error_message': str(e)
                }
                error_traceback = traceback.format_exc()
                print(error_traceback)

            if query_response.get('type') == SQL_RESPONSE_TYPE.ERROR:
                error_type = 'expected'
                error_code = query_response.get('error_code')
                error_text = query_response.get('error_message')

            context = mysql_proxy.get_context(context)

            query_response['context'] = context

        hooks.after_api_query(
            company_id=ctx.company_id,
            api='http',
            command=None,
            payload=query,
            error_type=error_type,
            error_code=error_code,
            error_text=error_text,
            traceback=error_traceback
        )

        return query_response, 200


@ns_conf.route('/list_databases')
@ns_conf.param('list_databases', 'lists databases of mindsdb')
class ListDatabases(Resource):
    @ns_conf.doc('list_databases')
    def get(self):
        listing_query = 'SHOW DATABASES'
        mysql_proxy = FakeMysqlProxy()
        try:
            result = mysql_proxy.process_query(listing_query)

            # iterate over result.data and perform a query on each item to get the name of the tables
            if result.type == SQL_RESPONSE_TYPE.ERROR:
                listing_query_response = {
                    'type': 'error',
                    'error_code': result.error_code,
                    'error_message': result.error_message
                }
            elif result.type == SQL_RESPONSE_TYPE.OK:
                listing_query_response = {
                    'type': 'ok'
                }
            elif result.type == SQL_RESPONSE_TYPE.TABLE:
                listing_query_response = {
                    'data': [{'name': x[0], 'tables': mysql_proxy.process_query('SHOW TABLES FROM `{}`'.format(x[0])).data} for x in result.data]
                }
        except Exception as e:
            listing_query_response = {
                'type': 'error',
                'error_code': 0,
                'error_message': str(e)
            }

        return listing_query_response, 200



import unittest
from mindsdb.integrations.handlers.nuo_jdbc_handler.nuo_jdbc_handler import NuoHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE

class NuoHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host": "localhost",
                "port": 48006,
                "database": "test",
                "schema": "hockey",
                "user": "dba",
                "password": "goalie",
                "is_direct": "true",
            }
        }
        cls.handler = NuoHandler('test_nuo_handler', **cls.kwargs)

    def test_0_connect(self):
        self.handler.connect()
    
    def test_1_check_connection(self):
        self.handler.check_connection()

    def test_2_create(self):
        res = self.handler.query('CREATE TABLE TESTTABLEX3 (ID INT PRIMARY KEY, NAME VARCHAR(14))')
        assert res.type is RESPONSE_TYPE.OK

    def test_3_insert(self):
        res = self.handler.query("INSERT INTO TESTTABLEX3 VALUES (100,'ONE HUNDRED'),(200,'TWO HUNDRED'),(300,'THREE HUNDRED')")
        assert res.type is RESPONSE_TYPE.OK

    def test_4_select(self):
        res = self.handler.query('SELECT * FROM HOCKEY')
        assert res.type is RESPONSE_TYPE.TABLE
    
    def test_5_get_tables(self):
        res = self.handler.get_tables()
        assert res.type is RESPONSE_TYPE.TABLE

    def test_6_get_columns(self):
        res = self.handler.get_columns("HOCKEY")
        assert res.type is RESPONSE_TYPE.TABLE
    

        
if __name__ == '__main__':
    unittest.main()


from ..mysql_handler import Handler as MySQLHandler
from .__about__ import __version__ as version

class SingleStoreHandler(MySQLHandler):
    """
    This handler handles connection and execution of the SingleStore statements.
    """
    name = 'singlestore'

    def __init__(self, name, **kwargs):
        kwargs['conn_attrs'] = {'mindsdb', 'MindsDB', version}
        super().__init__(name, **kwargs)



import ast
from collections import defaultdict
from typing import List

import nltk
import pandas as pd
from nltk import word_tokenize
from nltk.translate.bleu_score import (  # todo investigate why this always returns 0, not used for now
    sentence_bleu,
)
from nltk.translate.meteor_score import meteor_score
from rouge_score import rouge_scorer
from scipy.spatial import distance

from mindsdb.integrations.handlers.writer_handler.settings import (
    WriterHandlerParameters,
)
from mindsdb.utilities.log import get_log

# todo use polars for this for speed

logger = get_log(logger_name=__name__)


class WriterEvaluator:
    def __init__(self, args: WriterHandlerParameters, df: pd.DataFrame, rag):

        self.args = args
        self.df = df
        self.rag = rag(self.args)

        self.metric_map = {
            "cosine_similarity": self.calculate_cosine_similarities,
            "accuracy": self.get_matches,
            "rouge": self.calculate_rouge,
            "bleu": self.calculate_bleu,
            "meteor": self.calculate_meteor,
        }

        self.retrieval_metrics = self.args.retrieval_evaluation_metrics
        self.generator_metrics = self.args.generation_evaluation_metrics

        self.mean_evaluation_metrics = defaultdict(list)

        if args.evaluation_type == "e2e":
            # todo check if this is fine for cloud, better to download once and load from disk
            nltk.download("wordnet")

    def calculate_retrieval_metrics(
        self,
        df: pd.DataFrame,
        context_embeddings,
        retrieved_context_embeddings,
        prefix="retrieval_",
    ):
        """Calculate retrieval metrics"""

        for metric in self.retrieval_metrics:
            col_name = f"{prefix}{metric}"
            if metric == "cosine_similarity":
                df[col_name] = self.metric_map[metric](
                    context_embeddings, retrieved_context_embeddings
                )
            elif metric == "accuracy":
                col_name = f"{prefix}match"
                df[col_name] = self.get_matches(
                    gt_embeddings=context_embeddings,
                    test_embeddings=retrieved_context_embeddings,
                    threshold=self.args.retriever_match_threshold,
                )
            else:
                raise ValueError(f"metric {metric} not supported")

            self.store_mean_metric(col_name=col_name, mean_metric=df[col_name].mean())

        return df

    def calculate_generation_metrics(
        self,
        df: pd.DataFrame,
        generated_answer_embeddings,
        reference_answer_embeddings,
        prefix="generator_",
    ):
        """Calculate generation metrics"""

        for metric in self.generator_metrics:
            col_name = f"{prefix}{metric}"
            if metric == "cosine_similarity":

                df[col_name] = self.calculate_cosine_similarities(
                    generated_answer_embeddings, reference_answer_embeddings
                )
            elif metric == "accuracy":
                col_name = f"{prefix}match"
                df[col_name] = self.get_matches(
                    gt_embeddings=reference_answer_embeddings,
                    test_embeddings=generated_answer_embeddings,
                    threshold=self.args.generator_match_threshold,
                )
            elif metric == "rouge":
                df[col_name] = df.apply(
                    lambda x: self.metric_map[metric](
                        x["generated_answers"], x["reference_answers"]
                    ),
                    axis=1,
                )
                self.extract_rogue_scores(df, rogue_scores_col=col_name)
            elif metric == "bleu":
                df[col_name] = df.apply(
                    lambda x: self.metric_map[metric](
                        x["tokenized_generated_answers"],
                        x["tokenized_reference_answers"],
                    ),
                    axis=1,
                )
            elif metric == "meteor":
                df[col_name] = df.apply(
                    lambda x: self.calculate_meteor(
                        x["tokenized_generated_answers"],
                        x["tokenized_reference_answers"],
                    ),
                    axis=1,
                )
            else:
                raise ValueError(f"metric {metric} not supported")

            if metric != "rouge":

                self.store_mean_metric(
                    col_name=col_name, mean_metric=df[col_name].mean()
                )

        return df

    def embed_texts(self, texts: List[str]) -> List[list]:
        """Embed a list of texts"""
        return self.rag.embeddings_model.embed_documents(texts)

    def query_vector_store(self, question: str) -> List:
        """Query the vector store"""
        return self.rag.query_vector_store(question)

    @staticmethod
    def extract_returned_text(vector_store_response: List) -> List:
        # todo: this is a hack, we need to fix this so it works with multiple context ie top_k>1
        # todo handle empty response
        return [doc.page_content for doc in vector_store_response][0]

    def evaluation_prompt(self, question: str, context: str):
        """Create prompt for evaluating RAG"""

        if self.args.summarize_context:

            return self.rag.summarize_context(
                question=question, combined_context=context
            )

        return self.rag.prompt_template.format(question=question, context=context)

    def get_evaluation_prompts(self, df: pd.DataFrame) -> List[str]:
        """Create prompts for each question and context pair in the dataframe"""
        return [
            self.evaluation_prompt(question, context)
            for question, context in zip(df["question"], df["retrieved_context"])
        ]

    def extract_generated_texts(self, responses: List[str]):
        """Extract generated text from LLM response"""

        results = []
        for i, item in enumerate(responses):

            text = self.rag.extract_generated_text(item)

            results.append(text)
        return results

    @staticmethod
    def extract_reference_answers(df: pd.DataFrame) -> List[str]:
        """Get reference answers for each question in the dataframe"""

        # todo: this is a hack, we need to fix this so it works with multiple answers ie top_k>1
        answers = df["answers"].tolist()
        extracted_answers = []

        for answer in answers:
            try:
                extracted_answers.append(ast.literal_eval(answer)["text"][0])
            except IndexError as e:
                logger.error(e)
                extracted_answers.append("")
                continue

        return extracted_answers

    @staticmethod
    def _calculate_cosine_similarity(
        gt_embeddings: List[float], test_embeddings: List[float]
    ) -> float:
        """Calculate cosine similarity between a context and retrieved context embedding"""
        cosine_sim = 1 - distance.cosine(gt_embeddings, test_embeddings)

        return cosine_sim

    def calculate_cosine_similarities(
        self,
        gt_embeddings: List[List[float]],
        test_embeddings: List[List[float]],
    ):
        """Calculate cosine similarity for each ground truth and retrieved/generated pair for a given question"""

        return [
            self._calculate_cosine_similarity(
                context_embedding, retrieved_context_embedding
            )
            for context_embedding, retrieved_context_embedding in zip(
                gt_embeddings, test_embeddings
            )
        ]

    @staticmethod
    def check_match(cosine_similarity: float, threshold: float = 0.7) -> int:
        return int(cosine_similarity >= threshold)

    def get_matches(
        self, gt_embeddings, test_embeddings, threshold: float = 0.7
    ) -> List[int]:
        """Get matches for each ground truth and retrieved/generated pair for a given question"""

        cosine_similarities = self.calculate_cosine_similarities(
            gt_embeddings=gt_embeddings, test_embeddings=test_embeddings
        )

        matches = [
            self.check_match(cosine_similarity, threshold=threshold)
            for cosine_similarity in cosine_similarities
        ]

        return matches

    @staticmethod
    def _tokenize(text: str) -> List[str]:
        """Tokenize a text"""
        return word_tokenize(text)

    @staticmethod
    def calculate_rouge(generated: str, reference: str) -> dict:

        scorer = rouge_scorer.RougeScorer(["rouge1", "rougeL"], use_stemmer=True)
        score = scorer.score(generated, reference)
        return score

    def extract_rogue_scores(
        self, df: pd.DataFrame, rogue_scores_col: str = "rouge_scores"
    ):
        """Extract rouge scores from dataframe"""
        rouge_metrics = ["rouge1", "rougeL"]
        supported_metrics = ["precision", "recall", "fmeasure"]

        for rouge_metric in rouge_metrics:
            for supported_metric in supported_metrics:
                col_name = f"{rouge_metric}_{supported_metric}"
                df[col_name] = df.apply(
                    lambda x: getattr(
                        x[rogue_scores_col][rouge_metric], supported_metric
                    ),
                    axis=1,
                )

                self.store_mean_metric(
                    col_name=col_name, mean_metric=df[col_name].mean()
                )

    def store_mean_metric(self, col_name: str, mean_metric: float):
        """Calculate mean metric for each metric"""

        self.mean_evaluation_metrics[f"mean_{col_name}"].append(mean_metric)

    @staticmethod
    def calculate_bleu(
        generated_tokens: List[str], reference_tokens: List[str]
    ) -> float:
        return sentence_bleu([reference_tokens], generated_tokens)

    @staticmethod
    def calculate_meteor(
        generated_tokens: List[str], reference_tokens: List[str]
    ) -> float:
        return meteor_score([reference_tokens], generated_tokens)

    def evaluate_retrieval(self):
        """Evaluate the retrieval model"""

        df = self.df.copy(deep=True)

        # get question answering results
        df["retrieved_context"] = df.apply(
            lambda x: self.extract_returned_text(
                self.query_vector_store(x["question"])
            ),
            axis=1,
        )

        # embed context and retrieved context
        context_embeddings = self.embed_texts(df["context"].tolist())
        retrieved_context_embeddings = self.embed_texts(
            df["retrieved_context"].tolist()
        )

        df = self.calculate_retrieval_metrics(
            df, context_embeddings, retrieved_context_embeddings
        )

        return df

    def evaluate_generation(self, df: pd.DataFrame):
        """Evaluate the generation model, given the retrieval results df"""

        prompts = self.get_evaluation_prompts(df)

        raw_generated_answers = [self.rag.llm(prompt) for prompt in prompts]

        generated_answers = self.extract_generated_texts(raw_generated_answers)
        reference_answers = self.extract_reference_answers(df)

        df["generated_answers"] = generated_answers
        df["reference_answers"] = reference_answers

        # tokenize generated and reference answers
        df["tokenized_generated_answers"] = df.apply(
            lambda x: self._tokenize(x["generated_answers"]), axis=1
        )
        df["tokenized_reference_answers"] = df.apply(
            lambda x: self._tokenize(x["reference_answers"]), axis=1
        )

        # embed generated answers and reference answers
        generated_answer_embeddings = self.embed_texts(generated_answers)

        reference_answer_embeddings = self.embed_texts(reference_answers)

        df = self.calculate_generation_metrics(
            df, generated_answer_embeddings, reference_answer_embeddings
        )

        return df

    def evaluate_e2e(self):
        """Evaluate the end-to-end evaluation"""
        retrieval_df = self.evaluate_retrieval()
        e2e_df = self.evaluate_generation(retrieval_df)

        return e2e_df

    def evaluate(self):
        if self.args.evaluation_type == "retrieval":
            return self.evaluate_retrieval()
        elif self.args.evaluation_type == "e2e":
            return self.evaluate_e2e()
        else:
            raise ValueError(f"evaluation_type must be either 'retrieval' or 'e2e'")



from setuptools import find_packages, setup

pkgs, new_links = install_deps()

about = {}
with open("__about__.py") as fp:
    exec(fp.read(), about)

with open("requirements.txt") as req_file:
    requirements = [req.strip() for req in req_file.read().splitlines()]

setup(
    name=about["__title__"],
    version=about["__version__"],
    url=about["__github__"],
    download_url=about["__pypi__"],
    license=about["__license__"],
    author=about["__author__"],
    author_email=about["__email__"],
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=pkgs,
    dependency_links=new_links,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



from urllib.parse import urlparse, urljoin
import logging
import re
import traceback
from threading import Lock
import concurrent.futures
import io

import requests
from bs4 import BeautifulSoup
import pandas as pd
import fitz  # PyMuPDF


def pdf_to_markdown(response):
    # Download the PDF from the given URL
    
    file_stream = io.BytesIO(response.content)

    # Open the PDF from the in-memory file
    document = fitz.open(stream=file_stream, filetype='pdf')

    markdown_text = ''
    for page_num in range(len(document)):
        page = document.load_page(page_num)

        # Get the blocks of text
        blocks = page.get_text("blocks")

        # Sort the blocks by their vertical position on the page
        blocks.sort(key=lambda block: (block[1], block[0]))

        previous_block_bottom = 0
        for block in blocks:
            y0 = block[1]
            y1 = block[3]
            block_text = block[4]

            # Check if there's a large vertical gap between this block and the previous one
            if y0 - previous_block_bottom > 10:
                markdown_text += '\n'

            markdown_text += block_text + '\n'
            previous_block_bottom = y1

        markdown_text += '\n'

    # Close the document
    document.close()

    return markdown_text


url_list_lock = Lock()


def is_valid(url):
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)


# this bad boy gets all the crawling done in parallel
def parallel_get_all_website_links(urls):
    url_contents = {}

    if len(urls) <= 10:
        for url in urls:
            url_contents[url] = get_all_website_links(url)
        return url_contents
    
    with concurrent.futures.ProcessPoolExecutor() as executor:
        future_to_url = {executor.submit(get_all_website_links, url): url for url in urls}
        for future in concurrent.futures.as_completed(future_to_url):
            url = future_to_url[future]
            try:
                url_contents[url] = future.result()
            except Exception as exc:
                logging.error(f'{url} generated an exception: {exc}')
   
    return url_contents


# this crawls one individual website
def get_all_website_links(url):
    logging.info('crawling: {url} ...'.format(url=url))
    urls = set()
            
    domain_name = urlparse(url).netloc
    try:
        # Create a session to handle cookies
        session = requests.Session()

        # Add headers to mimic a real browser request
        headers = {
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/74.0.3729.169 Safari/537.36"
        }

        # Send GET request
        response = session.get(url, headers=headers)
        # Accept cookies if necessary
        if 'cookie' in response.request.headers:
            session.cookies.update(response.cookies)

        content_type = response.headers.get('Content-Type', '').lower()

        if 'application/pdf' in content_type:
            
            content_html = 'PDF'
            content_text = pdf_to_markdown(response)
        else:
            content_html = response.text

            # Parse HTML content with BeautifulSoup
            soup = BeautifulSoup(content_html, 'html.parser')
            content_text = get_readable_text_from_soup(soup)
            for a_tag in soup.findAll("a"):
                href = a_tag.attrs.get("href")
                if href == "" or href is None:
                    continue
                href = urljoin(url, href)
                parsed_href = urlparse(href)
                
                href = parsed_href.scheme + "://" + parsed_href.netloc + parsed_href.path
                if not is_valid(href):
                    continue
                if href in urls:
                    continue
                if domain_name != parsed_href.netloc:
                    continue
                
                href = href.rstrip('/')    
                urls.add(href)

    except Exception as e:
        error_message = traceback.format_exc().splitlines()[-1]
        logging.error("An exception occurred: %s", str(e))
        return {'url':url,'urls':urls, 'html_content':'', 'text_content': '', 'error':str(error_message)}

    return {'url': url,'urls': urls, 'html_content': content_html, 'text_content': content_text, 'error': None}


def get_readable_text_from_soup(soup):

    # Start formatting as Markdown
    markdown_output = ""

    # Iterate through headings and paragraphs
    for tag in soup.find_all(['h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'p', 'a', 'ul', 'ol', 'li']):
        if tag.name in ['h1', 'h2', 'h3', 'h4', 'h5', 'h6']:
            markdown_output += "#" * int(tag.name[1]) + " " + tag.get_text().strip() + "\n\n"
        elif tag.name == 'p':
            markdown_output += tag.get_text().strip() + "\n\n"
        elif tag.name == 'a':
            markdown_output += f"[{tag.get_text().strip()}]({tag.get('href')})\n\n"
        elif tag.name == 'ul':
            for li in tag.find_all('li'):
                markdown_output += f"* {li.get_text().strip()}\n"
            markdown_output += "\n"
        elif tag.name == 'ol':
            for index, li in enumerate(tag.find_all('li')):
                markdown_output += f"{index + 1}. {li.get_text().strip()}\n"
            markdown_output += "\n"

    return markdown_output


# this bad girl does the recursive crawling of the websites
def get_all_website_links_rec(url, reviewd_urls, limit=None):

    if limit is not None:
        if len(reviewd_urls) >= limit:
            return reviewd_urls

    if url not in reviewd_urls:
        # if something happens getting the website links for this url then log the error
        try:
            reviewd_urls[url] = get_all_website_links(url)
        except Exception as e:
            error_message = traceback.format_exc().splitlines()[-1]
            logging.error("An exception occurred: %s", str(e))
            reviewd_urls[url] = {'url': url, 'urls': [], 'html_content': '', 'text_content': '', 'error': str(error_message)}

    to_rev_url_list = []

    # create a list of new urls to review that don't exist in the already reviewed ones
    for new_url in reviewd_urls[url]['urls']:
        
        # if this is already in the urls, then no need to go and crawl for it
        if new_url in reviewd_urls or new_url in to_rev_url_list:
            continue
        
        # insert immediately to count limit between threads. fill later
        with url_list_lock:
            if limit is None or len(reviewd_urls) < limit:
                reviewd_urls[new_url] = {}
                to_rev_url_list.append(new_url)
            else:
                break

    # if there is something to fetch, go fetch
    if len(to_rev_url_list) > 0:
        new_revised_urls = parallel_get_all_website_links(to_rev_url_list)
    
        reviewd_urls.update(new_revised_urls)

        for new_url in new_revised_urls:
            get_all_website_links_rec(new_url, reviewd_urls, limit)


# this crawls the websites and retuns it all as a dataframe, ready to be served
def get_all_websites(urls, limit=1, html=False):
    reviewd_urls = {}

    # def fetch_url(url):
    #     url = url.rstrip('/')
    #     if urlparse(url).scheme == "":
    #         # Try HTTPS first
    #         url = "https://" + url
    #     reviewd_urls_iter = {}
    #     get_all_website_links_rec(url, reviewd_urls_iter, limit)
    #     return reviewd_urls_iter
    
    # reviewd_urls = fetch_url(urls[0])
    # Define a helper function that will be run in parallel.
    def fetch_url(url):
        # Allow URLs to be passed wrapped in quotation marks so they can be used
        # directly from the SQL editor.
        if url.startswith("'") and url.endswith("'"):
            url = url[1:-1]
        url = url.rstrip('/')
        if urlparse(url).scheme == "":
            # Try HTTPS first
            url = "https://" + url
        get_all_website_links_rec(url, reviewd_urls, limit)

    # Use a ThreadPoolExecutor to run the helper function in parallel.
    with concurrent.futures.ThreadPoolExecutor() as executor:
        future_to_url = {executor.submit(fetch_url, url): url for url in urls}

        for future in concurrent.futures.as_completed(future_to_url):
            future.result()

    columns_to_ignore = ['urls']
    if html is False:
        columns_to_ignore += ['html_content']
    df = dict_to_dataframe(reviewd_urls, columns_to_ignore=columns_to_ignore, index_name='url')

    if not df.empty and df[df.error.isna()].empty:
        # no real data - rise exception from first row
        raise Exception(str(df.iloc[0].error))
    return df


# this can parse the native query    
def parse_urls_limit(input_str):
    # Split the input string into 'url', 'limit' or 'html' parts
    items = re.split(r',', input_str)

    # Initialize list for urls, limit and html
    urls = []
    limit = None
    html = False

    for item in items:
        item = item.strip()  # Remove leading/trailing whitespace

        # Check if item is a 'limit' or 'html' setting
        if item.lower().startswith('limit'):
            limit_match = re.search(r'\d+', item)
            if limit_match:
                limit = int(limit_match.group())  # Extract the number
        elif item.lower().startswith('html'):
            html_match = re.search(r'(true|false)', item, re.I)
            if html_match:
                html = html_match.group().lower() == 'true'  # Check if the value is 'true'
        else:
            urls.append(item)  # Add the item to the url list

    return {"urls": urls, "limit": limit, "html": html}


# run a query that goes and crawls urls
# format url, url, ..., limit=n
# you can pass one of many urls, limit is optional
def get_df_from_query_str(query_str):
    args = parse_urls_limit(query_str)
    df = get_all_websites(args['urls'], args['limit'], args['html'])
    return df


# this flips a dictionary of dictionaries into a dataframe so we can use it in mindsdb
def dict_to_dataframe(dict_of_dicts, columns_to_ignore=None, index_name=None):
    # Convert dictionary of dictionaries into DataFrame
    df = pd.DataFrame.from_dict(dict_of_dicts, orient='index')

    # If columns_to_ignore is provided, drop these columns
    if columns_to_ignore:
        df = df.drop(columns_to_ignore, axis=1, errors='ignore')

    # If index_name is provided, rename the index
    if index_name:
        df.index.name = index_name

    return df


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    response = requests.get('https://www.goldmansachs.com/media-relations/press-releases/current/pdfs/2023-q2-results.pdf')
    print(pdf_to_markdown(response))




import pandas as pd
from typing import Text, List
from mindsdb_sql.parser import ast
from abc import ABC, abstractmethod
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class BaseQueryParser(ABC):
    """
    Parses a SQL query into its component parts.

    Parameters
    ----------
    query : ast
        Given SQL query.
    """

    def __init__(self, query: ast):
        self.query = query

    @abstractmethod
    def parse_query(self):
        """
        Parses a SQL statement into its components.
        """
        pass

    def parse_where_clause(self) -> List[List[Text]]:
        """
        Parses the WHERE clause of the query.
        """
        where_conditions = extract_comparison_conditions(self.query.where)
        return where_conditions
    

class BaseQueryExecutor():
    """
    Executes a SQL query.

    Parameters
    ----------
    query : ast
        Given SQL query.
    """

    def __init__(self, df: pd.DataFrame, where_conditions: List[List[Text]]):
        self.df = df
        self.where_conditions = where_conditions

    def execute_query(self):
        """
        Executes the SQL query.
        """
        self.execute_where_clause()

        return self.df

    def execute_where_clause(self):
        """
        Executes the where clause of the query.
        """
        if len(self.where_conditions) > 0:
            for condition in self.where_conditions:
                column = condition[1]
                operator = '==' if condition[0] == '=' else condition[0]
                value = f"'{condition[2]}'" if type(condition[2]) == str else condition[2]

                query = f"{column} {operator} {value}"
                self.df.query(query, inplace=True)



from typing import Optional
from collections import OrderedDict

import pandas as pd
import sqlite3

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.base import DatabaseHandler

from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class SQLiteHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the SQLite statements.
    """

    name = 'sqlite'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.parser = parse_sql
        self.dialect = 'sqlite'
        self.connection_data = connection_data
        self.kwargs = kwargs

        self.connection = None
        self.is_connected = False

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection

        self.connection = sqlite3.connect(self.connection_data['db_file'])
        self.is_connected = True

        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to SQLite {self.connection_data["db_file"]}, {e}!')
            response.error_message = str(e)
        finally:
            if response.success is True and need_to_close:
                self.disconnect()
            if response.success is False and self.is_connected is True:
                self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """
        Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        cursor = connection.cursor()

        try:
            cursor.execute(query)
            result = cursor.fetchall()
            if result:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    data_frame=pd.DataFrame(
                        result,
                        columns=[x[0] for x in cursor.description]
                    )
                )
            else:
                connection.commit()
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            log.logger.error(f'Error running query: {query} on {self.connection_data["db_file"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )

        cursor.close()
        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> StatusResponse:
        """
        Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind
                of query: SELECT, INTSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        renderer = SqlalchemyRender('sqlite')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> StatusResponse:
        """
        Return list of entities that will be accessible as tables.
        Returns:
            HandlerResponse
        """

        query = "SELECT name from sqlite_master where type= 'table';"
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name: str) -> StatusResponse:
        """
        Returns a list of entity columns.
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """

        query = f"PRAGMA table_info([{table_name}]);"
        result = self.native_query(query)
        df = result.data_frame
        result.data_frame = df.rename(columns={'name': 'column_name', 'type': 'data_type'})
        return result


connection_args = OrderedDict(
    db_file={
        'type': ARG_TYPE.STR,
        'description': 'The database file where the data will be stored. The special path name :memory: can be provided'
                       ' to create a temporary database in RAM.'
    }
)

connection_args_example = OrderedDict(
    db_file='chinook.db'
)


import json
from shutil import copyfile
from collections import OrderedDict

import requests

from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response
)

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE

from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions
from mindsdb.integrations.libs.api_handler import APIHandler, APITable
from mindsdb_sql.parser import ast
from mindsdb.utilities import log
from mindsdb_sql import parse_sql
from mindsdb.utilities.config import Config

import os
import time
from typing import List
import pandas as pd

from google.auth.transport.requests import Request
from google.oauth2.credentials import Credentials
from google_auth_oauthlib.flow import Flow
from googleapiclient.discovery import build
from googleapiclient.errors import HttpError
from email.message import EmailMessage

from base64 import urlsafe_b64encode, urlsafe_b64decode

from .utils import AuthException, google_auth_flow, save_creds_to_file

DEFAULT_SCOPES = ['https://www.googleapis.com/auth/gmail.compose',
                  'https://www.googleapis.com/auth/gmail.readonly',
                  'https://www.googleapis.com/auth/gmail.modify']




class EmailsTable(APITable):
    """Implementation for the emails table for Gmail"""

    def select(self, query: ast.Select) -> pd.DataFrame:
        """Pulls emails from Gmail "users.messages.list" API

        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query

        Returns
        -------
        pd.DataFrame
            Email matching the query

        Raises
        ------
        NotImplementedError
            If the query contains an unsupported operation or condition
        """

        conditions = extract_comparison_conditions(query.where)

        params = {}
        include_attachments = False
        for op, arg1, arg2 in conditions:

            if op == 'or':
                raise NotImplementedError(f'OR is not supported')

            if arg1 in ['query', 'label_ids', 'include_spam_trash', 'include_attachments']:
                if op == '=':
                    if arg1 == 'query':
                        params['q'] = arg2
                    elif arg1 == 'label_ids':
                        params['labelIds'] = arg2.split(',')
                    elif arg1 == 'include_attachments':
                        include_attachments = arg2 == 'true'
                    else:
                        params['includeSpamTrash'] = arg2
                else:
                    raise NotImplementedError(f'Unknown op: {op}')

            else:
                raise NotImplementedError(f'Unknown clause: {arg1}')

        if query.limit is not None:
            params['maxResults'] = query.limit.value

        result = self.handler.call_gmail_api(
            method_name='list_messages',
            params=params
        )
        attachments = []
        if include_attachments:
            attachments = self.handler.get_attachments(result)
        # filter targets
        columns = []
        for target in query.targets:
            if isinstance(target, ast.Star):
                columns = self.get_columns()
                break
            elif isinstance(target, ast.Identifier):
                columns.append(target.parts[-1])
            else:
                raise NotImplementedError(f"Unknown query target {type(target)}")

        # columns to lower case
        columns = [name.lower() for name in columns]

        if len(result) == 0:
            return pd.DataFrame([], columns=columns)

        # add absent columns
        for col in set(columns) & set(result.columns) ^ set(columns):
            result[col] = None

        # filter by columns
        result = result[columns]

        # Rename columns
        for target in query.targets:
            if target.alias:
                result.rename(columns={target.parts[-1]: str(target.alias)}, inplace=True)

        return result

    def get_columns(self) -> List[str]:
        """Gets all columns to be returned in pandas DataFrame responses

        Returns
        -------
        List[str]
            List of columns
        """
        return [
            'id',
            'message_id',
            'thread_id',
            'label_ids',
            'sender',
            'to',
            'date',
            'subject',
            'snippet',
            'history_id',
            'size_estimate',
            'body',
            'attachments',
        ]

    def insert(self, query: ast.Insert):
        """Sends reply emails using the Gmail "users.messages.send" API

        Parameters
        ----------
        query : ast.Insert
           Given SQL INSERT query

        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """
        columns = [col.name for col in query.columns]

        supported_columns = {"message_id", "thread_id", "to_email", "subject", "body"}
        if not set(columns).issubset(supported_columns):
            unsupported_columns = set(columns).difference(supported_columns)
            raise ValueError(
                "Unsupported columns for create email: "
                + ", ".join(unsupported_columns)
            )

        for row in query.values:
            params = dict(zip(columns, row))

            if not 'to_email' in params:
                raise ValueError('"to_email" parameter is required to send an email')

            message = EmailMessage()
            message['To'] = params['to_email']
            message['Subject'] = params['subject'] if 'subject' in params else ''

            content = params['body'] if 'body' in params else ''
            message.set_content(content)

            # If threadId is present then add References and In-Reply-To headers
            # so that proper threading can happen
            if 'thread_id' in params and 'message_id' in params:
                message['In-Reply-To'] = params['message_id']
                message['References'] = params['message_id']

            encoded_message = urlsafe_b64encode(message.as_bytes()).decode()

            message = {
                'raw': encoded_message
            }

            if 'thread_id' in params:
                message['threadId'] = params['thread_id']
            self.handler.call_gmail_api('send_message', {'body': message})

    def delete(self, query: ast.Delete):
        """
        Deletes an event or events in the calendar.

        Args:
            query (ast.Delete): SQL query to parse.

        Returns:
            Response: Response object containing the results.
        """

        # Parse the query to get the conditions.
        conditions = extract_comparison_conditions(query.where)
        for op, arg1, arg2 in conditions:
            if op == 'or':
                raise NotImplementedError(f'OR is not supported')
            if arg1 == 'message_id':
                if op == '=':
                    self.handler.call_gmail_api('delete_message', {'id': arg2})
                else:
                    raise NotImplementedError(f'Unknown op: {op}')
            else:
                raise NotImplementedError(f'Unknown clause: {arg1}')

    def update(self, query: ast.Update) -> None:
        """Updates a label of a message.

        Args:
            query (ASTNode): The SQL query to parse.

        Raises:
            NotImplementedError: If the query contains an unsupported condition.
        """
        params = {}
        conditions = extract_comparison_conditions(query.where)
        for op, arg1, arg2 in conditions:
            if op == 'or':
                raise NotImplementedError(f'OR is not supported')
            if arg1 == 'id':
                if op == '=':
                    params['id'] = arg2
                else:
                    raise NotImplementedError(f'Unknown op: {op}')
            else:
                raise NotImplementedError(f'Unknown clause: {arg1}')
        request_body = {}
        values = query.update_columns.items()
        data_list = list(values)
        add_label = []
        remove_label = []
        for key, value in data_list:
            if key == 'addLabel':
                add_label.append(str(value)[1:-1])
            elif key == 'removeLabel':
                remove_label.append(str(value)[1:-1])
            else:
                raise NotImplementedError(f'Unknown clause: {key}')
        if add_label:
            request_body['addLabelIds'] = add_label
        if remove_label:
            request_body['removeLabelIds'] = remove_label
        params['body'] = request_body
        self.handler.call_gmail_api('update_message', params)


class GmailHandler(APIHandler):
    """A class for handling connections and interactions with the Gmail API.

    Attributes:
        credentials_file (str): The path to the Google Auth Credentials file for authentication
        and interacting with the Gmail API on behalf of the uesr.

        scopes (List[str], Optional): The scopes to use when authenticating with the Gmail API.
    """

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.connection_args = kwargs.get('connection_data', {})

        self.credentials_url = self.connection_args.get('credentials_url', None)
        self.credentials_file = self.connection_args.get('credentials_file', None)
        if self.connection_args.get('credentials'):
            self.credentials_file = self.connection_args.pop('credentials')
        if not self.credentials_file and not self.credentials_url:
            # try to get from config
            gm_config = Config().get('handlers', {}).get('gmail', {})
            secret_file = gm_config.get('credentials_file')
            secret_url = gm_config.get('credentials_url')
            if secret_file:
                self.credentials_file = secret_file
            elif secret_url:
                self.credentials_url = secret_url

        self.scopes = self.connection_args.get('scopes', DEFAULT_SCOPES)
        self.token_file = None
        self.max_page_size = 500
        self.max_batch_size = 100
        self.service = None
        self.is_connected = False

        self.handler_storage = kwargs['handler_storage']

        emails = EmailsTable(self)
        self.emails = emails
        self._register_table('emails', emails)

    def _download_secret_file(self, secret_file):
        # Giving more priority to the S3 file
        if self.credentials_url:
            response = requests.get(self.credentials_url)
            if response.status_code == 200:
                with open(secret_file, 'w') as creds:
                    creds.write(response.text)
                return True
            else:
                log.logger.error("Failed to get credentials from S3", response.status_code)

        if self.credentials_file and os.path.isfile(self.credentials_file):
            copyfile(self.credentials_file, secret_file)
            return True
        return False

    def create_connection(self):
        creds = None

        # Get the current dir, we'll check for Token & Creds files in this dir
        curr_dir = self.handler_storage.folder_get('config')

        creds_file = os.path.join(curr_dir, 'creds.json')
        secret_file = os.path.join(curr_dir, 'secret.json')

        if os.path.isfile(creds_file):
            creds = Credentials.from_authorized_user_file(creds_file, self.scopes)

        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())

            if self._download_secret_file(secret_file):
                # save to storage
                self.handler_storage.folder_sync('config')
            else:
                raise ValueError('No valid Gmail Credentials filepath or S3 url found.')

            creds = google_auth_flow(secret_file, self.scopes, self.connection_args.get('code'))

            save_creds_to_file(creds, creds_file)
            self.handler_storage.folder_sync('config')

        return build('gmail', 'v1', credentials=creds)

    def connect(self):
        """Authenticate with the Gmail API using the credentials file.

        Returns
        -------
        service: object
            The authenticated Gmail API service object.
        """
        if self.is_connected and self.service is not None:
            return self.service

        self.service = self.create_connection()

        self.is_connected = True
        return self.service

    def check_connection(self) -> StatusResponse:
        """Check connection to the handler.

        Returns
        -------
        StatusResponse
            Status confirmation
        """
        response = StatusResponse(False)

        try:
            # Call the Gmail API
            service = self.connect()

            result = service.users().getProfile(userId='me').execute()

            if result and result.get('emailAddress', None) is not None:
                response.success = True
        except AuthException as error:
            response.error_message = str(error)
            response.redirect_url = error.auth_url
            return response

        except HttpError as error:
            response.error_message = f'Error connecting to Gmail api: {error}.'
            log.logger.error(response.error_message)

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query_string: str = None) -> Response:
        ast = parse_sql(query_string, dialect="mindsdb")

        return self.query(ast)

    def _parse_parts(self, parts, attachments):
        if not parts:
            return ''

        body = ''
        for part in parts:
            if part['mimeType'] == 'text/plain':
                part_body = part.get('body', {}).get('data', '')
                body += urlsafe_b64decode(part_body).decode('utf-8')
            elif part['mimeType'] == 'multipart/alternative' or 'parts' in part:
                # Recursively iterate over nested parts to find the plain text body
                body += self._parse_parts(part['parts'], attachments)
            elif part.get('filename') and part.get('body') and part.get('body').get('attachmentId'):
                # For now just store the attachment details
                attachments.append({
                    'filename': part['filename'],
                    'mimeType': part['mimeType'],
                    'attachmentId': part['body']['attachmentId']
                })
            else:
                log.logger.debug(f"Unhandled mimeType: {part['mimeType']}")

        return body

    def _parse_message(self, data, message, exception):
        if exception:
            log.logger.error(f'Exception in getting full email: {exception}')
            return

        payload = message['payload']
        headers = payload.get("headers", [])
        parts = payload.get("parts")

        row = {
            'id': message['id'],
            'thread_id': message['threadId'],
            'label_ids': message.get('labelIds', []),
            'snippet': message.get('snippet', ''),
            'history_id': message['historyId'],
            'size_estimate': message.get('sizeEstimate', 0),
        }

        for header in headers:
            key = header['name'].lower()
            value = header['value']

            if key in ['to', 'subject', 'date']:
                row[key] = value
            elif key == 'from':
                row['sender'] = value
            elif key == 'message-id':
                row['message_id'] = value

        attachments = []
        row['body'] = self._parse_parts(parts, attachments)
        row['attachments'] = json.dumps(attachments)
        data.append(row)

    def _get_messages(self, data, messages):
        batch_req = self.service.new_batch_http_request(
            lambda id, response, exception: self._parse_message(data, response, exception))
        for message in messages:
            batch_req.add(self.service.users().messages().get(userId='me', id=message['id']))

        batch_req.execute()

    def get_attachments(self, result):
        for index, email in result.iterrows():
            attachments = json.loads(email['attachments'])
            for attachment in attachments:
                attachment_id = attachment['attachmentId']
                filename = attachment['filename']
                mimeType = attachment['mimeType']
                attachment_data = self.service.users().messages().attachments().get(
                    userId='me', messageId=email['id'], id=attachment_id).execute()
                file_data = attachment_data['data']
                file_data = file_data.replace('-', '+').replace('_', '/')
                file_data = urlsafe_b64decode(file_data)
                with open(filename, 'wb') as f:
                    f.write(file_data)

    def _handle_list_messages_response(self, data, messages):
        total_pages = len(messages) // self.max_batch_size
        for page in range(total_pages):
            self._get_messages(data, messages[page * self.max_batch_size:(page + 1) * self.max_batch_size])

        # Get the remaining messsages, if any
        if len(messages) % self.max_batch_size > 0:
            self._get_messages(data, messages[total_pages * self.max_batch_size:])

    def call_gmail_api(self, method_name: str = None, params: dict = None) -> pd.DataFrame:
        """Call Gmail API and map the data to pandas DataFrame
        Args:
            method_name (str): method name
            params (dict): query parameters
        Returns:
            DataFrame
        """
        service = self.connect()
        if method_name == 'list_messages':
            method = service.users().messages().list
        elif method_name == 'send_message':
            method = service.users().messages().send
        elif method_name == "delete_message":
            method = service.users().messages().trash
        elif method_name == 'update_message':
            method = service.users().messages().modify
        else:
            raise NotImplementedError(f'Unknown method_name: {method_name}')

        left = None
        count_results = None
        if 'maxResults' in params:
            count_results = params['maxResults']

        params['userId'] = 'me'

        data = []
        limit_exec_time = time.time() + 60

        while True:
            if time.time() > limit_exec_time:
                raise RuntimeError('Handler request timeout error')

            if count_results is not None:
                left = count_results - len(data)
                if left == 0:
                    break
                elif left < 0:
                    # got more results that we need
                    data = data[:left]
                    break

                if left > self.max_page_size:
                    params['maxResults'] = self.max_page_size
                else:
                    params['maxResults'] = left

            log.logger.debug(f'Calling Gmail API: {method_name} with params ({params})')

            resp = method(**params).execute()

            if 'messages' in resp:
                self._handle_list_messages_response(data, resp['messages'])
            elif isinstance(resp, dict):
                data.append(resp)

            if count_results is not None and 'nextPageToken' in resp:
                params['pageToken'] = resp['nextPageToken']
            else:
                break

        df = pd.DataFrame(data)

        return df


connection_args = OrderedDict(
    credentials_url={
        'type': ARG_TYPE.STR,
        'description': 'URL to Service Account Keys',
        'label': 'URL to Service Account Keys',
    },
    credentials_file={
        'type': ARG_TYPE.STR,
        'description': 'Location of Service Account Keys',
        'label': 'path of Service Account Keys',
    },
    credentials={
        'type': ARG_TYPE.PATH,
        'description': 'Service Account Keys',
        'label': 'Upload Service Account Keys',
    },
)



import pandas as pd
from mindsdb.integrations.libs.api_handler import APITable
from mindsdb_sql.parser import ast
from mindsdb.integrations.utilities.sql_utils import extract_comparison_conditions


class AccountsTable(APITable):

    def flatten_dict(self, data: dict, prefix: str = ""):
        flat_data = {}
        for key, value in data.items():
            if isinstance(value, dict):
                flattened_sub_dict = self.flatten_dict(value, f"{key}_")
                flat_data.update(flattened_sub_dict)
            else:
                flat_data[f"{prefix}{key}"] = value
        return flat_data

    def select(self, query: ast.Select) -> pd.DataFrame:
        qbo = self.handler.connect()
        accounts_data = qbo.accounts.get()
        flattened_accounts_data = [self.flatten_dict(account) for account in accounts_data]
        result = pd.DataFrame(flattened_accounts_data)
        self.filter_columns(result, query)
        return result

    def get_columns(self):
        return [
            'Name',
            'SubAccount',
            'FullyQualifiedName',
            'Active',
            'Classification',
            'AccountType',
            'AccountSubType',
            'CurrentBalance',
            'CurrentBalanceWithSubAccounts',
            'CurrencyRef_value',
            'CurrencyRef_name',
            'domain',
            'sparse',
            'Id',
            'SyncToken',
            'MetaData_CreateTime',
            'MetaData_LastUpdatedTime',
        ]

    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None):
        columns = []
        if query is not None:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    columns.append(target.value)
        if len(columns) > 0:
            result = result[columns]


class PurchasesTable(APITable):

    def flatten_dict(self, data: dict, prefix: str = ""):
        flat_data = {}
        for key, value in data.items():
            if isinstance(value, dict):
                flattened_sub_dict = self.flatten_dict(value, f"{key}_")
                flat_data.update(flattened_sub_dict)
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, dict):
                        flattened_sub_dict = self.flatten_dict(item, f"{key}_{i}_")
                        flat_data.update(flattened_sub_dict)
                    else:
                        flat_data[f"{prefix}{key}_{i}"] = item
            else:
                flat_data[f"{prefix}{key}"] = value
        return flat_data



    def select(self, query: ast.Select) -> pd.DataFrame:
        qbo = self.handler.connect()
        purchases_data = qbo.purchases.get()
        flattened_purchases_data = [self.flatten_dict(purchase) for purchase in purchases_data]
        result = pd.DataFrame(flattened_purchases_data)
        self.filter_columns(result, query)
        return result

    def get_columns(self):
        return [
            'AccountRef_value',
            'AccountRef_name',
            'PaymentType',
            'Credit',
            'TotalAmt',
            'domain',
            'sparse',
            'Id',
            'SyncToken',
            'MetaData_CreateTime',
            'MetaData_LastUpdatedTime',
            'TxnDate',
            'CurrencyRef_value',
            'CurrencyRef_name',
            'EntityRef_value',
            'EntityRef_name',
            'EntityRef_type',
            'Line_0_Id',
            'Line_0_Amount',
            'Line_0_DetailType'
            # Add more columns for additional line items if needed
        ]

    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None):
        columns = []
        if query is not None:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    columns.append(target.value)
        if len(columns) > 0:
            result = result[columns]

class BillPaymentsTable(APITable):
    def flatten_dict(self, data: dict, prefix: str = ""):
        flat_data = {}
        for key, value in data.items():
            if isinstance(value, dict):
                flattened_sub_dict = self.flatten_dict(value, f"{prefix}{key}_")
                flat_data.update(flattened_sub_dict)
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, dict):
                        flattened_sub_dict = self.flatten_dict(item, f"{prefix}{key}_{i}_")
                        flat_data.update(flattened_sub_dict)
                    else:
                        flat_data[f"{prefix}{key}_{i}"] = item
            else:
                flat_data[prefix + key] = value
        return flat_data

    def select(self, query: ast.Select) -> pd.DataFrame:
        qbo = self.handler.connect()
        billpayments_data = qbo.bill_payments.get()
        flattened_billpayments_data = [self.flatten_dict(bp) for bp in billpayments_data]
        result = pd.DataFrame(flattened_billpayments_data)
        self.filter_columns(result, query)
        return result

    def get_columns(self):
        return [
            'VendorRef_value',
            'VendorRef_name',
            'PayType',
            'CreditCardPayment_CCAccountRef_value',
            'CreditCardPayment_CCAccountRef_name',
            'CheckPayment_BankAccountRef_value',
            'CheckPayment_BankAccountRef_name',
            'TotalAmt',
            'Id',
            'SyncToken',
            'MetaData_CreateTime',
            'MetaData_LastUpdatedTime',
            'DocNumber',
            'TxnDate',
            'CurrencyRef_value',
            'CurrencyRef_name',
            'Line_0_Amount',
            'Line_0_LinkedTxn_0_TxnId',
            'Line_0_LinkedTxn_0_TxnType',
            'domain',
            'sparse',
        ]
        
    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None):
        columns = []
        if query is not None:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    columns.append(target.value)
        if len(columns) > 0:
            result = result[columns]

class VendorsTable(APITable):
    
    def flatten_dict(self, data: dict, prefix: str = ""):
        flat_data = {}
        for key, value in data.items():
            if isinstance(value, dict):
                flattened_sub_dict = self.flatten_dict(value, f"{key}_")
                flat_data.update(flattened_sub_dict)
            else:
                flat_data[f"{prefix}{key}"] = value
        return flat_data
    
    def select(self, query: ast.Select) -> pd.DataFrame:
        qbo = self.handler.connect()
        vendors_data = qbo.vendors.get()
        flattened_vendors_data = [self.flatten_dict(vendor) for vendor in vendors_data]
        result = pd.DataFrame(flattened_vendors_data)
        self.filter_columns(result, query)
        return result
    
    def get_columns(self):
        return [
            'Balance',
            'Vendor1099',
            'CurrencyRef_value',
            'CurrencyRef_name',
            'Id',
            'SyncToken',
            'MetaData_CreateTime',
            'MetaData_LastUpdatedTime',
            'DisplayName',
            'PrintOnCheckName',
            'Active',
            'domain',
            'sparse'
        ]
    
    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None):
        columns = []
        if query is not None:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    columns.append(target.value)
        if len(columns) > 0:
            result = result[columns]


class BillsTable(APITable):

    def flatten_dict(self, data: dict, prefix: str = ""):
        flat_data = {}
        for key, value in data.items():
            if isinstance(value, dict):
                flattened_sub_dict = self.flatten_dict(value, f"{prefix}{key}_")
                flat_data.update(flattened_sub_dict)
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, dict):
                        flattened_sub_dict = self.flatten_dict(item, f"{prefix}{key}_{i}_")
                        flat_data.update(flattened_sub_dict)
                    else:
                        flat_data[f"{prefix}{key}_{i}"] = item
            else:
                flat_data[prefix + key] = value
        return flat_data

    def select(self, query: ast.Select) -> pd.DataFrame:
        qbo = self.handler.connect()
        bills_data = qbo.bills.get()
        flattened_bills_data = [self.flatten_dict(bill) for bill in bills_data]
        result = pd.DataFrame(flattened_bills_data)
        self.filter_columns(result, query)
        return result

    def get_columns(self):
        return [
            'DueDate',
            'Balance',
            'Id',
            'SyncToken',
            'MetaData_CreateTime',
            'MetaData_LastUpdatedTime',
            'TxnDate',
            'CurrencyRef_value',
            'CurrencyRef_name',
            'VendorRef_value',
            'VendorRef_name',
            'APAccountRef_value',
            'APAccountRef_name',
            'TotalAmt',
            'Line_0_Id',
            'Line_0_LineNum',
            'Line_0_Description',
            'Line_0_Amount',
            'Line_0_DetailType',
            'Line_0_ItemBasedExpenseLineDetail_BillableStatus',
            'Line_0_ItemBasedExpenseLineDetail_ItemRef_value',
            'Line_0_ItemBasedExpenseLineDetail_ItemRef_name',
            'Line_0_ItemBasedExpenseLineDetail_UnitPrice',
            'Line_0_ItemBasedExpenseLineDetail_Qty',
            'Line_1_Id',
            'Line_1_LineNum',
            'Line_1_Description',
            'Line_1_Amount',
            'Line_1_DetailType',
            'Line_1_ItemBasedExpenseLineDetail_BillableStatus',
            'Line_1_ItemBasedExpenseLineDetail_ItemRef_value',
            'Line_1_ItemBasedExpenseLineDetail_ItemRef_name',
            'Line_1_ItemBasedExpenseLineDetail_UnitPrice',
            'Line_1_ItemBasedExpenseLineDetail_Qty',
            'domain',
            'sparse'
        ]

    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None):
        columns = []
        if query is not None:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    columns.append(target.value)
        if len(columns) > 0:
            result = result[columns]

class EmployeesTable(APITable):
    
    def flatten_dict(self, data: dict, prefix: str = ""):
        flat_data = {}
        for key, value in data.items():
            if isinstance(value, dict):
                flattened_sub_dict = self.flatten_dict(value, f"{prefix}{key}_")
                flat_data.update(flattened_sub_dict)
            elif isinstance(value, list):
                for i, item in enumerate(value):
                    if isinstance(item, dict):
                        flattened_sub_dict = self.flatten_dict(item, f"{prefix}{key}_{i}_")
                        flat_data.update(flattened_sub_dict)
                    else:
                        flat_data[f"{prefix}{key}_{i}"] = item
            else:
                flat_data[prefix + key] = value
        return flat_data

    def select(self, query: ast.Select) -> pd.DataFrame:
        qbo = self.handler.connect()
        employees_data = qbo.employees.get()
        flattened_employees_data = [self.flatten_dict(employee) for employee in employees_data]
        result = pd.DataFrame(flattened_employees_data)
        self.filter_columns(result, query)
        return result

    def get_columns(self):
        return [
            'BillableTime',
            'Id',
            'SyncToken',
            'MetaData_CreateTime',
            'MetaData_LastUpdatedTime',
            'GivenName',
            'FamilyName',
            'DisplayName',
            'PrintOnCheckName',
            'Active',
            'PrimaryPhone_FreeFormNumber',
            'HiredDate',
            'domain',
            'sparse'
        ]

    def filter_columns(self, result: pd.DataFrame, query: ast.Select = None):
        columns = []
        if query is not None:
            for target in query.targets:
                if isinstance(target, ast.Star):
                    columns = self.get_columns()
                    break
                elif isinstance(target, ast.Identifier):
                    columns.append(target.value)
        if len(columns) > 0:
            result = result[columns]


from typing import List


class ANSWER_TYPE:
    __slots__ = ()
    TABLE = "table"
    OK = "ok"
    ERROR = "error"


ANSWER_TYPE = ANSWER_TYPE()


class ExecuteAnswer:
    def __init__(
        self,
        # TODO remove, is not used
        answer_type: ANSWER_TYPE,
        columns=None,
        data: List[List] = None,
        status: int = None,
        state_track: List[List] = None,
        error_code: int = None,
        error_message: str = None,
    ):
        self.columns = columns
        self.data = data
        self.status = status
        self.state_track = state_track
        self.error_code = error_code
        self.error_message = error_message



import pandas as pd

from solace.messaging.messaging_service import MessagingService, RetryStrategy
from solace.messaging.resources.topic_subscription import TopicSubscription
from solace.messaging.receiver.message_receiver import MessageHandler
from solace.messaging.resources.topic import Topic
from solace.messaging.receiver.inbound_message import InboundMessage

from mindsdb_sql import parse_sql
from mindsdb_sql.parser import ast

from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
)
from mindsdb.utilities import log


class SolaceHandler(DatabaseHandler):

    def __init__(self, name: str = None, **kwargs):
        """Registers all API tables and prepares the handler for an API connection.

        Args:
            name: (str): The handler name to use
        """
        super().__init__(name)
        args = kwargs.get('connection_data', {})

        if 'host' not in args:
            raise ValueError('Host parameter is required')

        self.connection_args = args
        self.messaging_service = None
        self.is_connected = False

    def connect(self):

        broker_props = {
            "solace.messaging.transport.host": self.connection_args['host'],
            "solace.messaging.service.vpn-name": self.connection_args.get('vpn-name', 'default'),
            "solace.messaging.authentication.scheme.basic.username": self.connection_args.get('username'),
            "solace.messaging.authentication.scheme.basic.password": self.connection_args.get('password')
        }

        self.messaging_service = MessagingService\
            .builder()\
            .from_properties(broker_props) \
            .with_reconnection_retry_strategy(RetryStrategy.parametrized_retry(20, 3)) \
            .build()

        self.messaging_service.connect()

        self.direct_publisher = self.messaging_service\
            .create_direct_message_publisher_builder()\
            .build()
        self.direct_publisher.start()

        # TODO support persistent_publisher ?

        self.is_connected = True
        return self.messaging_service

    def disconnect(self):

        if self.is_connected is False:
            return

        self.direct_publisher.terminate()
        self.messaging_service.disconnect()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:

        response = StatusResponse(False)

        try:
            self.connect()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to Solace: {e}!')
            response.error_message = e

        if response.success is False:
            self.is_connected = False
        return response

    def native_query(self, query: str = None) -> Response:
        ast = parse_sql(query, dialect='mindsdb')
        return self.query(ast)

    def query(self, query: ast.ASTNode):
        if isinstance(query, ast.Insert):
            result = self.handle_insert(query)
        else:
            raise NotImplementedError
        return result

    def get_columns(self, table_name: str) -> Response:
        df = pd.DataFrame([], columns=['Field'])
        df['Type'] = 'str'

        return Response(RESPONSE_TYPE.TABLE, df)

    def get_tables(self) -> Response:
        df = pd.DataFrame([], columns=['table_name', 'table_type'])

        return Response(RESPONSE_TYPE.TABLE, df)

    def handle_insert(self, query: ast.Insert):

        message_builder = self.messaging_service.message_builder()

        topic_name = '/'.join(query.table.parts)

        column_names = [col.name for col in query.columns]
        for insert_row in query.values:
            data = dict(zip(column_names, insert_row))

            outbound_message = message_builder.build(data)
            self.direct_publisher.publish(destination=Topic.of(topic_name), message=outbound_message)

        return Response(RESPONSE_TYPE.OK)

    def subscribe(self, stop_event, callback, table_name, columns=None, **kwargs):

        class MessageHandlerImpl(MessageHandler):
            def on_message(self, message: 'InboundMessage'):
                # Check if the payload is a dict
                payload = message.get_payload_as_dictionary()
                if payload is not None:
                    data = payload
                else:
                    # check as string
                    payload = message.get_payload_as_string()
                    if payload is None:
                        payload = message.get_payload_as_bytes()
                        if isinstance(payload, bytearray):
                            payload = payload.decode()
                    data = {'data': payload}

                if columns is not None:
                    updated_columns = data.keys()
                    if not set(columns) & set(updated_columns):
                        # skip
                        return

                callback(data)

        table = ast.Identifier(table_name)
        topic_name = '/'.join(table.parts)

        topics = [TopicSubscription.of(topic_name)]
        direct_receiver = self.messaging_service\
            .create_direct_message_receiver_builder()\
            .with_subscriptions(topics)\
            .build()

        direct_receiver.start()
        direct_receiver.receive_async(MessageHandlerImpl())

        stop_event.wait()

        direct_receiver.terminate()



import setuptools

about = {}
with open("__about__.py") as fp:
    exec(fp.read(), about)

with open('requirements.txt') as req_file:
    requirements = [req.strip() for req in req_file.read().splitlines()]

setuptools.setup(
    name=about['__title__'],
    version=about['__version__'],
    url=about['__github__'],
    download_url=about['__pypi__'],
    license=about['__license__'],
    author=about['__author__'],
    description=about['__description__'],
    packages=setuptools.find_packages(),
    package_data={'project': ['requirements.txt']},
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8"
)



import os
import tempfile
from collections import OrderedDict

import pandas as pd
import requests

from cassandra.cluster import Cluster
from cassandra.auth import PlainTextAuthProvider
from cassandra.util import Date

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser import ast
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender

from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log


class ScyllaHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Scylla statements.
    """
    name = 'scylla'

    def __init__(self, name=None, **kwargs):
        super().__init__(name)
        self.parser = parse_sql
        self.connection_args = kwargs.get('connection_data')
        self.session = None
        self.is_connected = False

    def download_secure_bundle(self, url, max_size=10*1024*1024):
        """
        Downloads the secure bundle from a given URL and stores it in a temporary file.
        
        :param url: URL of the secure bundle to be downloaded.
        :param max_size: Maximum allowable size of the bundle in bytes. Defaults to 10MB.
        :return: Path to the downloaded secure bundle saved as a temporary file.
        :raises ValueError: If the secure bundle size exceeds the allowed `max_size`.
        
        TODO:
        - Find a way to periodically clean up or delete the temporary files 
        after they have been used to prevent filling up storage over time.
        """
        response = requests.get(url, stream=True, timeout=10)
        response.raise_for_status()

        content_length = int(response.headers.get('content-length', 0))
        if content_length > max_size:
            raise ValueError("Secure bundle is larger than the allowed size!")

        with tempfile.NamedTemporaryFile(delete=False) as temp_file:
            size_downloaded = 0
            for chunk in response.iter_content(chunk_size=8192):
                size_downloaded += len(chunk)
                if size_downloaded > max_size:
                    raise ValueError("Secure bundle is larger than the allowed size!")
                temp_file.write(chunk)
        return temp_file.name

    def connect(self):
        """
        Handles the connection to a Scylla keystore.
        """
        if self.is_connected is True:
            return self.session

        auth_provider = PlainTextAuthProvider(
            username=self.connection_args['user'], password=self.connection_args['password']
        )

        connection_props = {
            'auth_provider': auth_provider
        }
        connection_props['protocol_version'] = self.connection_args.get('protocol_version', 4)
        secure_connect_bundle = self.connection_args.get('secure_connect_bundle')

        if secure_connect_bundle:
            # Check if the secure bundle is a URL
            if secure_connect_bundle.startswith(('http://', 'https://')):
                secure_connect_bundle = self.download_secure_bundle(secure_connect_bundle)
            connection_props['cloud'] = {
                'secure_connect_bundle': secure_connect_bundle
            }
        else:
            connection_props['contact_points'] = [self.connection_args['host']]
            connection_props['port'] = int(self.connection_args['port'])

        cluster = Cluster(**connection_props)
        session = cluster.connect(self.connection_args.get('keyspace'))

        self.is_connected = True
        self.session = session
        return self.session

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the Scylla database
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)

        try:
            session = self.connect()
            # TODO: change the healthcheck
            session.execute('SELECT release_version FROM system.local').one()
            response.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to Scylla {self.connection_args["keyspace"]}, {e}!')
            response.error_message = e

        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def prepare_response(self, resp):
        # replace cassandra types
        data = []
        for row in resp:
            row2 = {}
            for k, v in row._asdict().items():
                if isinstance(v, Date):
                    v = v.date()
                row2[k] = v
            data.append(row2)
        return data

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in MySQL
        :return: returns the records from the current recordset
        """
        session = self.connect()
        try:
            resp = session.execute(query).all()
            resp = self.prepare_response(resp)
            if resp:
                response = Response(
                    RESPONSE_TYPE.TABLE,
                    pd.DataFrame(
                        resp
                    )
                )
            else:
                response = Response(RESPONSE_TYPE.OK)
        except Exception as e:
            log.logger.error(f'Error running query: {query} on {self.connection_args["keyspace"]}!')
            response = Response(
                RESPONSE_TYPE.ERROR,
                error_message=str(e)
            )
        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """

        # remove table alias because Cassandra Query Language doesn't support it
        if isinstance(query, ast.Select):
            if isinstance(query.from_table, ast.Identifier) and query.from_table.alias is not None:
                query.from_table.alias = None

            # remove table name from fields
            table_name = query.from_table.parts[-1]

            for target in query.targets:
                if isinstance(target, ast.Identifier):
                    if target.parts[0] == table_name:
                        target.parts.pop(0)

        renderer = SqlalchemyRender('mysql')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in MySQL
        """
        q = "DESCRIBE TABLES;"
        result = self.native_query(q)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"DESCRIBE {table_name};"
        result = self.native_query(q)
        return result


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'User name',
        'required': True,
        'label': 'User'
    },
    password={
        'type': ARG_TYPE.PWD,
        'description': 'Password',
        'required': True,
        'label': 'Password'
    },
    protocol_version={
        'type': ARG_TYPE.INT,
        'description': 'is not required, and default to 4.',
        'required': False,
        'label': 'Protocol version'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': ' is the host name or IP address of the ScyllaDB.',
        'required': True,
        'label': 'Host'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Server port',
        'required': True,
        'label': 'Port'
    },
    keyspace={
        'type': ARG_TYPE.STR,
        'description': 'is the keyspace to connect to. It is a top level container for tables.',
        'required': True,
        'label': 'Keyspace'
    },
    secure_connect_bundle={
        'type': ARG_TYPE.STR,
        'description': 'Path or URL to the secure connect bundle',
        'required': True,
        'label': 'Host'
    }
)



import os
import sys
import logging
import traceback

from mindsdb.utilities.config import Config
from functools import partial


class LoggerWrapper(object):
    def __init__(self, writer_arr, default_writer_pos):
        self._writer_arr = writer_arr
        self.default_writer_pos = default_writer_pos

    def write(self, message):
        if len(message.strip(' \n')) == 0:
            return
        if 'DEBUG:' in message:
            self._writer_arr[0](message)
        elif 'INFO:' in message:
            self._writer_arr[1](message)
        elif 'WARNING:' in message:
            self._writer_arr[2](message)
        elif 'ERROR:' in message:
            self._writer_arr[3](message)
        else:
            self._writer_arr[self.default_writer_pos](message)

    def flush(self):
        pass

    def isatty(self):
        return True  # assumes terminal attachment

    def fileno(self):
        return 1  # stdout

# class DbHandler(logging.Handler):
#     def __init__(self):
#         logging.Handler.__init__(self)
#         self.company_id = os.environ.get('MINDSDB_COMPANY_ID', None)
#
#     def emit(self, record):
#         self.format(record)
#         if (
#             len(record.message.strip(' \n')) == 0
#             or (record.threadName == 'ray_print_logs' and 'mindsdb-logger' not in record.message)
#         ):
#             return
#
#         log_type = record.levelname
#         source = f'file: {record.pathname} - line: {record.lineno}'
#         payload = record.msg
#
#         if telemtry_enabled:
#             pass
#             # @TODO: Enable once we are sure no sensitive info is being outputed in the logs
#             # if log_type in ['INFO']:
#             #    add_breadcrumb(
#             #        category='auth',
#             #        message=str(payload),
#             #        level='info',
#             #    )
#             # Might be too much traffic if we send this for users with slow networks
#             # if log_type in ['DEBUG']:
#             #    add_breadcrumb(
#             #        category='auth',
#             #        message=str(payload),
#             #        level='debug',
#             #    )
#
#         if log_type in ['ERROR', 'WARNING']:
#             trace = str(traceback.format_stack(limit=20))
#             trac_log = Log(log_type='traceback', source=source, payload=trace, company_id=self.company_id)
#             session.add(trac_log)
#             session.commit()
#
#             if telemtry_enabled:
#                 add_breadcrumb(
#                     category='stack_trace',
#                     message=trace,
#                     level='info',
#                 )
#                 if log_type in ['ERROR']:
#                     capture_message(str(payload))
#                 if log_type in ['WARNING']:
#                     capture_message(str(payload))
#
#         log = Log(log_type=str(log_type), source=source, payload=str(payload), company_id=self.company_id)
#         session.add(log)
#         session.commit()

# default logger
logger = logging.getLogger('dummy')


def initialize_log(config=None, logger_name='main', wrap_print=False):
    global logger
    if config is None:
        config = Config().get_all()

    telemtry_enabled = os.getenv('CHECK_FOR_UPDATES', '1').lower() not in ['0', 'false', 'False']

    if telemtry_enabled:
        try:
            import sentry_sdk
            from sentry_sdk import capture_message, add_breadcrumb
            sentry_sdk.init(
                "https://29e64dbdf325404ebf95473d5f4a54d3@o404567.ingest.sentry.io/5633566",
                traces_sample_rate=0  # Set to `1` to experiment with performance metrics
            )
        except (ImportError, ModuleNotFoundError) as e:
            raise Exception(f"to use telemetry please install 'pip install mindsdb[telemetry]': {e}")

    ''' Create new logger
    :param config: object, app config
    :param logger_name: str, name of logger
    :param wrap_print: bool, if true, then print() calls will be wrapped by log.debug() function.
    '''
    log = logging.getLogger(f'mindsdb.{logger_name}')
    log.propagate = False
    log.setLevel(min(
        getattr(logging, config['log']['level']['console']),
        getattr(logging, config['log']['level']['file'])
    ))

    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')

    console_handler = logging.StreamHandler()
    console_handler.setLevel(config['log']['level'].get('console', logging.INFO))
    console_handler.setFormatter(formatter)
    log.addHandler(console_handler)

    # db_handler = DbHandler()
    # db_handler.setLevel(config['log']['level'].get('db', logging.WARNING))
    # db_handler.setFormatter(formatter)
    # log.addHandler(db_handler)

    if wrap_print:
        sys.stdout = LoggerWrapper([log.debug, log.info, log.warning, log.error], 1)
        sys.stderr = LoggerWrapper([log.debug, log.info, log.warning, log.error], 3)

    log.error = partial(log.error, exc_info=True)
    logger = log


def get_log(logger_name=None):
    if logger_name is None:
        return logging.getLogger('mindsdb')
    return logging.getLogger(f'mindsdb.{logger_name}')




import requests
from datetime import datetime
from typing import Dict, Optional

import pandas as pd
from mlflow.tracking import MlflowClient

from mindsdb.integrations.libs.base import BaseMLEngine


class MLflowHandler(BaseMLEngine):
    """
    The MLflow integration engine needs to have a working connection to MLFlow. For this:
        - All models to use should be previously served
        - An MLflow server should be running, to access its model registry

    Example:
        1. Run `mlflow server -p 5001 --backend-store-uri sqlite:////path/to/mlflow.db --default-artifact-root ./artifacts --host 0.0.0.0`
        2. Run `mlflow models serve --model-uri ./model_path`
        3. Run MindsDB
    
    Note: above, `artifacts` is a folder to store artifacts for new experiments that do not specify an artifact store.
    """  # noqa

    name = 'mlflow'

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        args = args['using']  # ignore the rest of the problem definition
        connection = MlflowClient(args['mlflow_server_url'], args['mlflow_server_path'])
        model_name = args['model_name']
        mlflow_models = [model.name for model in connection.search_registered_models()]

        if model_name not in mlflow_models:
            raise Exception(f"Error: model '{model_name}' not found in mlflow. Check serving and try again.")

        args['target'] = target
        self._check_model_url(args['predict_url'])
        self.model_storage.json_set('args', args)

    def predict(self, df, args=None):
        args = self.model_storage.json_get('args')  # override any incoming args for now
        self._check_model_url(args['predict_url'])
        resp = requests.post(args['predict_url'],
                             data=df.to_json(orient='records'),
                             headers={'content-type': 'application/json; format=pandas-records'})
        answer = resp.json()
        predictions = pd.DataFrame({args['target']: answer})
        return predictions

    def describe(self, key: Optional[str] = None) -> pd.DataFrame:
        if key == 'info':
            args = self.model_storage.json_get('args')
            connection = MlflowClient(args['mlflow_server_url'], args['self.mlflow_server_path'])
            models = {model.name: model for model in connection.search_registered_models()}
            model = models[key]
            latest_version = model.latest_versions[-1]
            description = {
                'NAME': [model.name],
                'USER_DESCRIPTION': [model.description],
                'LAST_STATUS': [latest_version.status],
                'CREATED_AT': [datetime.fromtimestamp(model.creation_timestamp//1000).strftime("%m/%d/%Y, %H:%M:%S")],
                'LAST_UPDATED': [datetime.fromtimestamp(model.last_updated_timestamp//1000).strftime("%m/%d/%Y, %H:%M:%S")],
                'TAGS': [model.tags],
                'LAST_RUN_ID': [latest_version.run_id],
                'LAST_SOURCE_PATH': [latest_version.source],
                'LAST_USER_ID': [latest_version.user_id],
                'LAST_VERSION': [latest_version.version],
            }
            return pd.DataFrame.from_dict(description)
        else:
            tables = ['info']
            return pd.DataFrame(tables, columns=['tables'])


    @staticmethod
    def _check_model_url(url):
        """ try post without data, check status code not in (not_found, method_not_allowed) """
        try:
            resp = requests.post(url)
            if resp.status_code in (404, 405):
                raise Exception(f'Model url is incorrect, status_code: {resp.status_code}')
        except requests.RequestException as e:
            raise Exception(f'Model url is incorrect: {str(e)}')
